# Module containing useful function to interact with sep files
import re
import os
import numpy as np
from occamypy.utils.os import RunShellCmd

# Assigning datapath
HOME = os.environ["HOME"]
datapath = None
# Checking environment definition first
if "DATAPATH" in os.environ:
    datapath = os.environ["DATAPATH"]
# Checking local directory
elif os.path.isfile('.datapath'):
    out = (RunShellCmd("cat .datapath | head -n 1", check_code=False, get_stat=False)[0]).rstrip()
    datapath = out.split("=")[1]
# Checking whether the local host has a datapath
else:
    if os.path.isfile(HOME + "/.datapath"):
        out = RunShellCmd("cat $HOME/.datapath | grep $HOST", check_code=False, get_stat=False)[0]
        if len(out) == 0:
            out = (RunShellCmd("cat $HOME/.datapath | head -n 1", check_code=False, get_stat=False)[0]).rstrip()
        datapath = out.split("=")[1]

# Checking if datapath was found
if datapath is None:
    if os.path.isdir("/tmp/"):
        datapath = "/tmp/"
        print("WARNING! DATAPATH not found. The folder /tmp will be used to write binary files")
    else:
        raise IOError("SEP datapath not found\n Set env variable DATAPATH to a folder to write binary files")


def rm_file(filename):
    """File to remove header and binary files"""
    binfile = get_binary(filename)
    if os.path.isfile(filename):
        os.remove(filename)
    if os.path.isfile(binfile):
        os.remove(binfile)
    return


def get_par(filename, par):
    """ Function to obtain a header parameter within the passed header file"""
    info = None
    # Checking if label is requested
    if "label" in par:
        reg_prog = re.compile("%s=(\'(.*?)\'|\"(.*?)\")" % par)
    else:
        reg_prog = re.compile("%s=([^\s]+)" % par)
    if not os.path.isfile(filename):
        raise OSError("ERROR! No %s file found!" % filename)
    for line in reversed(open(filename).readlines()):
        if info is None:
            find = reg_prog.search(line)
            if find:
                info = find.group(1)
    if info is None:
        raise IOError("%s parameter not found in file %s" % (filename, par))
    # Removing extra characters from found parameter
    if info is not None:
        info = info.replace('"', '')
        info = info.replace('\'', '')
    return info


def get_binary(filename):
    """ Function to obtain binary file associated with a given header file"""
    return get_par(filename, "in")


def get_axes(filename):
    """Function returning all axis information related to a header file"""
    axes = []
    # Currently handling maximum 7 axis
    for iaxis in range(7):
        # Obtaining number of elements within each axis
        try:
            axis_n = int(get_par(filename, par="n%s" % (iaxis + 1)))
        except IOError as exc:
            if iaxis == 0:
                print(exc.args)
                print("ERROR! First axis parameters must be found! Returning None")
                return None
            else:
                # Default value for an unset axis
                axis_n = 1
        # Obtaining origin of each axis
        try:
            axis_o = float(get_par(filename, par="o%s" % (iaxis + 1)))
        except IOError as exc:
            if iaxis == 0:
                print(exc.args)
                print("ERROR! First axis parameters must be found! Returning None")
                return None
            else:
                # Default value for an unset axis
                axis_o = 0.0
        # Obtaining sampling of each axis
        try:
            axis_d = float(get_par(filename, par="d%s" % (iaxis + 1)))
        except IOError as exc:
            if iaxis == 0:
                print(exc.args)
                print("ERROR! First axis parameters must be found! Returning None")
                return None
            else:
                # Default value for an unset axis
                axis_d = 0.0
        # Obtaining label of each axis
        try:
            axis_lab = get_par(filename, par="label%s" % (iaxis + 1))
        except IOError as exc:
            # Default value for an unset axis
            axis_lab = "Undefined"
        axes.append([axis_n, axis_o, axis_d, axis_lab])
    return axes


def get_num_axes(filename):
    """Function to obtain number of axes in a header file"""
    # Obtaining elements in each dimensions
    axis_info = get_axes(filename)
    axis_elements = [ii[0] for ii in axis_info]
    index = [i for i, nelements in enumerate(axis_elements) if int(nelements) > 1]
    if index:
        n_axes = index[-1] + 1
    else:
        n_axes = 1
    return n_axes


def read_file(filename, formatting='>f', mem_order="C"):
    """Function for reading header files"""
    axis_info = get_axes(filename)
    n_axis = get_num_axes(filename)
    shape = [ii[0] for ii in axis_info]
    shape = shape[:n_axis]
    if mem_order == "C":
        shape = tuple(reversed(shape))
    elif mem_order != "F":
        raise ValueError("ERROR! %s not an supported array order" % mem_order)
    fid = open(get_binary(filename), 'r+b')
    # Default formatting big-ending floating point number
    data = np.fromfile(fid, dtype=formatting)
    # Reshaping array and forcing memory continuity
    if mem_order == "C":
        data = np.ascontiguousarray(np.reshape(data, shape, order=mem_order))
    else:
        data = np.asfortranarray(np.reshape(data, shape, order=mem_order))
    fid.close()
    return [data, axis_info]


def write_file(filename, data, axis_info=None, formatting='>f'):
    """Function for writing header files"""
    global datapath
    # write binary file
    binfile = datapath + filename.split('/')[-1] + '@'
    with open(binfile, 'w+b') as fid:
        # Default formatting big-ending floating point number
        if np.isfortran(data):  # Forcing column-wise binary writing
            data.flatten('F').astype(formatting).tofile(fid)
        else:
            data.astype(formatting).tofile(fid)
    fid.close()
    # If axis_info is not provided all the present axis are set to d=1.0 o=0.0 label='Undefined'
    if axis_info is None:
        naxis = data.shape
        if not np.isfortran(data):
            naxis = tuple(reversed(naxis))  # If C last axis is the "fastest"
        axis_info = [[naxis[ii], 0.0, 1.0, 'Undefined'] for ii in range(0, len(naxis))]
    # writing header/pointer file
    with open(filename, 'w') as fid:
        # Writing axis info
        for ii, ax_info in enumerate(axis_info):
            ax_id = ii + 1
            fid.write("n%s=%s o%s=%s d%s=%s label%s='%s'\n"
                      % (ax_id, ax_info[0], ax_id, ax_info[1], ax_id, ax_info[2], ax_id, ax_info[3]))
        fid.write("in='%s'\n" % binfile)
        fid.write("data_format='xdr_float'\n")
        fid.write("esize=4\n")
    fid.close()
    return



import numpy as np
from math import isnan
from copy import deepcopy


class Stepper:
    """Stepper parent object"""

    # Default class methods/functions
    def __init__(self):
        """Default class constructor for Stepper"""
        return

    def __del__(self):
        """Default destructor"""
        return

    def run(self, model, search_dir):
        """Dummy stepper running method"""
        raise NotImplementedError("Implement run stepper in the derived class.")

    def estimate_initial_guess(self, problem, modl, dmodl, logger):
        """Function to estimate initial step length value"""
        try:
            # Projecting search direction in the data space
            dres = problem.get_dres(modl, dmodl)
        except NotImplementedError:
            if logger:
                logger.addToLog(
                    "\t!!!dresf not implemented; stepper will use inverse of search direction norm as initial step length value!!!")
            # Initial step length value of 1.0 / |dm|_2
            alpha_guess = 1.0 / dmodl.norm()
            return alpha_guess
        res = problem.get_res(modl)
        dres_res = res.dot(dres)
        dres_dres = dres.dot(dres)
        if dres_dres == 0.:
            if logger:
                logger.addToLog(
                    "\t!!!Gradient in the null space of linear forward operator; using inverse of search direction norm as step length value!!!")
            # Initial step length value of 1.0 / |dm|_2
            alpha_guess = 1.0 / dmodl.norm()
        else:
            # alpha = -phi'(0)/phi''(0)
            alpha_guess = -dres_res / dres_dres
        return alpha_guess


class CvSrchStep(Stepper):
    """
    Originally published by More and Thuente (1994) "Line Search Algorithms with Guaranteed Sufficient Decrease"
    CvSrch stepper (from Dianne O'Leary's code):

    THE PURPOSE OF CVSRCH IS TO FIND A STEP WHICH SATISFIES
    A SUFFICIENT DECREASE CONDITION AND A CURVATURE CONDITION.

    AT EACH STAGE THE SUBROUTINE UPDATES AN INTERVAL OF
    UNCERTAINTY WITH ENDPOINTS STX AND STY. THE INTERVAL OF
    UNCERTAINTY IS INITIALLY CHOSEN SO THAT IT CONTAINS A
    MINIMIZER OF THE MODIFIED FUNCTION

        F(X+STP*S) - F(X) - FTOL*STP*(GRADF(X)'S).

    IF A STEP IS OBTAINED FOR WHICH THE MODIFIED FUNCTION
    HAS A NONPOSITIVE FUNCTION VALUE AND NONNEGATIVE DERIVATIVE,
    THEN THE INTERVAL OF UNCERTAINTY IS CHOSEN SO THAT IT
    CONTAINS A MINIMIZER OF F(X+STP*S).

    THE ALGORITHM IS DESIGNED TO FIND A STEP WHICH SATISFIES
    THE SUFFICIENT DECREASE CONDITION

         F(X+STP*S) .LE. F(X) + FTOL*STP*(GRADF(X)'S),

    AND THE CURVATURE CONDITION

         ABS(GRADF(X+STP*S)'S)) .LE. GTOL*ABS(GRADF(X)'S).

    IF FTOL IS LESS THAN GTOL AND IF, FOR EXAMPLE, THE FUNCTION
    IS BOUNDED BELOW, THEN THERE IS ALWAYS A STEP WHICH SATISFIES
    BOTH CONDITIONS. IF NO STEP CAN BE FOUND WHICH SATISFIES BOTH
    CONDITIONS, THEN THE ALGORITHM USUALLY STOPS WHEN ROUNDING
    ERRORS PREVENT FURTHER PROGRESS. IN THIS CASE STP ONLY
    SATISFIES THE SUFFICIENT DECREASE CONDITION.
    """

    def __init__(self, alpha=0.0, xtol=1.0e-16, ftol=1.0e-4, gtol=0.95, alpha_min=1.0e-20, alpha_max=1.e20, maxfev=20,
                 xtrapf=4., delta=0.66):
        """
           CvSrch constructor:
           alpha 		 = [0.] - float; Initial step-length guess
           xtol  	 	 = [1e-16] - float; Relative width tolerance: convergence is reached if width falls below xtol * maximum step size.
           ftol  	 	 = [1e-16] - float; c1 value to tests first Wolfe condition (should be between 0 and 1)
           gtol  	 	 = [0.95] - float; c2 value to tests second Wolfe condition (should be between c1 or ftol and 1). For Quasi-Newton (e.g., L-BFGS) choose default. Otherwise, for other methods (e.g., NLCG) choose 0.1
           alpha_min  	 = [1e-20] - float; Minimum step length value of the step length interval
           alpha_max  	 = [1e20] - float; Maximum step length value of the step length interval
           maxfev  	     = [20] - int; Maximum number of function evaluation to step length
           xtrapf  	     = [4.0] - float; Scaling factor to find right limit of uncertainty interval
           delta  	     = [0.66] - float; Value to force sufficient decrease of interval size on successive iterations. Should be a positive value less than 1.
        """
        self.alpha = alpha  # Initial step length guess
        self.xtol = xtol
        self.ftol = ftol
        self.gtol = gtol
        self.alpha_min = alpha_min
        self.alpha_max = alpha_max
        self.maxfev = maxfev
        self.xtrapf = xtrapf
        self.delta = delta
        self.zero = 10 ** (np.floor(
            np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow
        # Checking stepper parameters
        if xtol < 0.:
            raise ValueError("ERROR! xtol must be greater than 0.0, current value %.2e" % xtol)
        if ftol < 0. or ftol >= 1.:
            raise ValueError("ERROR! ftol must be greater than 0.0 and smaller than 1.0, current value %.2e" % ftol)
        if gtol < 0. or ftol >= 1.:
            raise ValueError("ERROR! gtol must be greater than 0.0 and smaller than 1.0, current value %.2e" % gtol)
        if alpha_min < 0.:
            raise ValueError("ERROR! alpha_min must be greater than 0.0, current value %.2e" % alpha_min)
        if xtrapf < 0.:
            raise ValueError("ERROR! xtrapf must be greater than 0.0, current value %.2e" % xtrapf)
        if delta < 0. or delta >= 1.:
            raise ValueError("ERROR! delta must be greater than 0.0 and smaller than 1.0, current value %.2e" % delta)
        if maxfev < 0:
            raise ValueError("ERROR! maxfev must be greater than 0, current value %d" % maxfev)
        if alpha_max < alpha_min:
            raise ValueError(
                "ERROR! alpha_max must be greater than alpha_min, current values: alpha_min=%.2e; alpha_max=%.2e"
                % (alpha_min, alpha_max))

    def cstep(self, stx, fx, dx, sty, fy, dy, stp, fp, dp, brackt, stpmin, stpmax, logger):
        """
            Modified Cstep function (from the code by Dianne O'Leary July 1991):
            The purpose of cstep is to compute a safeguarded step for
            a linesearch and to update an interval of uncertainty for
            a minimizer of the function.

            The parameter stx contains the step with the least function
            value. The parameter stp contains the current step. It is
            assumed that the derivative at stx is negative in the
            direction of the step. If brackt is set true then a
            minimizer has been bracketed in an interval of uncertainty
            with endpoints stx and sty.
            The subroutine statement is

            subroutine cstep(stx,fx,dx,sty,fy,dy,stp,fp,dp,brackt,stpmin,stpmax,info)

            where

            stx, fx, and dx are variables which specify the step,
            the function, and the derivative at the best step obtained
            so far. The derivative must be negative in the direction
            of the step, that is, dx and stp-stx must have opposite
            signs. On output these parameters are updated appropriately.

            sty, fy, and dy are variables which specify the step,
            the function, and the derivative at the other endpoint of
            the interval of uncertainty. On output these parameters are
            updated appropriately.

            stp, fp, and dp are variables which specify the step,
            the function, and the derivative at the current step.
            If brackt is set true then on input stp must be
            between stx and sty. On output stp is set to the new step.

            brackt is a logical variable which specifies if a minimizer
            has been bracketed. If the minimizer has not been bracketed
            then on input brackt must be set false. If the minimizer
            is bracketed then on output brackt is set true.

            stpmin and stpmax are input variables which specify lower
            and upper bounds for the step.

            info is an integer output variable set as follows:
            If info = True, then the step has been computed
            according to one of the five cases below. Otherwise
            info = False, and this indicates improper input parameters.
        """

        success = False  # which is info

        # Check the input parameters for errors.
        if ((brackt and (stp <= np.minimum(stx, sty) or
                         stp >= np.maximum(stx, sty))) or
                dx * (stp - stx) >= 0.0 or
                stpmax < stpmin):
            if logger:
                logger.addToLog("\tFunction cstep could find step and update interval of uncertainty!")
            return stx, fx, dx, sty, fy, dy, stp, fp, dp, brackt, success

        # Determine if the derivatives have opposite sign.
        sgnd = dp * (dx / np.abs(dx))

        # First case. A higher function value.
        # The minimum is bracketed. If the cubic step is closer
        # to stx than the quadratic step, the cubic step is taken,
        # else the average of the cubic and quadratic steps is taken.

        if fp > fx:
            success = True
            bound = True
            theta = 3.0 * (fx - fp) / (stp - stx) + dx + dp
            s = np.linalg.norm([theta, dx, dp], np.inf)
            gamma = s * np.sqrt((theta / s) * (theta / s) - (dx / s) * (dp / s));
            if stp < stx:
                gamma = -gamma
            p = (gamma - dx) + theta
            q = ((gamma - dx) + gamma) + dp
            r = p / q
            stpc = stx + r * (stp - stx)
            stpq = stx + ((dx / ((fx - fp) / (stp - stx) + dx)) / 2.0) * (stp - stx)
            if np.abs(stpc - stx) < np.abs(stpq - stx):
                stpf = stpc
            else:
                stpf = stpc + (stpq - stpc) / 2.0
            brackt = True

        # Second case. A lower function value and derivatives of
        # opposite sign. The minimum is bracketed. If the cubic
        # step is closer to stx than the quadratic (secant) step,
        # the cubic step is taken, else the quadratic step is taken.

        elif sgnd < 0.0:
            success = True
            bound = False
            theta = 3.0 * (fx - fp) / (stp - stx) + dx + dp
            s = np.linalg.norm([theta, dx, dp], np.inf)
            gamma = s * np.sqrt((theta / s) * (theta / s) - (dx / s) * (dp / s))
            if stp > stx:
                gamma = -gamma
            p = (gamma - dp) + theta
            q = ((gamma - dp) + gamma) + dx
            r = p / q
            stpc = stp + r * (stx - stp)
            stpq = stp + (dp / (dp - dx)) * (stx - stp)
            if np.abs(stpc - stp) > np.abs(stpq - stp):
                stpf = stpc
            else:
                stpf = stpq
            brackt = True

        # Third case. A lower function value, derivatives of the
        # same sign, and the magnitude of the derivative decreases.
        # The cubic step is only used if the cubic tends to infinity
        # in the direction of the step or if the minimum of the cubic
        # is beyond stp. Otherwise the cubic step is defined to be
        # either stpmin or stpmax. The quadratic (secant) step is also
        # computed and if the minimum is bracketed then the the step
        # closest to stx is taken, else the step farthest away is taken.

        elif np.abs(dp) < np.abs(dx):
            success = True
            bound = True
            theta = 3.0 * (fx - fp) / (stp - stx) + dx + dp
            s = np.linalg.norm([theta, dx, dp], np.inf)

            # The case gamma = 0 only arises if the cubic does not tend
            # to infinity in the direction of the step.

            gamma = s * np.sqrt(np.maximum(0., (theta / s) * (theta / s) - (dx / s) * (dp / s)))
            if stp > stx:
                gamma = -gamma

            p = (gamma - dp) + theta
            q = (gamma + (dx - dp)) + gamma
            r = p / q
            if r < 0.0 and gamma != 0.0:
                stpc = stp + r * (stx - stp)
            elif stp > stx:
                stpc = stpmax
            else:
                stpc = stpmin
            stpq = stp + (dp / (dp - dx)) * (stx - stp)
            if brackt:
                if np.abs(stp - stpc) < np.abs(stp - stpq):
                    stpf = stpc
                else:
                    stpf = stpq
            else:
                if np.abs(stp - stpc) > np.abs(stp - stpq):
                    stpf = stpc
                else:
                    stpf = stpq

        # Fourth case. A lower function value, derivatives of the
        # same sign, and the magnitude of the derivative does
        # not decrease. If the minimum is not bracketed, the step
        # is either stpmin or stpmax, else the cubic step is taken.

        else:
            success = True
            bound = False
            if brackt:
                theta = 3.0 * (fp - fy) / (sty - stp) + dy + dp
                s = np.linalg.norm([theta, dx, dp], np.inf)
                gamma = s * np.sqrt((theta / s) * (theta / s) - (dy / s) * (dp / s))
                if stp > sty:
                    gamma = -gamma
                p = (gamma - dp) + theta
                q = ((gamma - dp) + gamma) + dy
                r = p / q
                stpc = stp + r * (sty - stp)
                stpf = stpc
            elif stp > stx:
                stpf = stpmax
            else:
                stpf = stpmin

        # Update the interval of uncertainty. This update does not
        # depend on the new step or the case analysis above.

        if fp > fx:
            sty = stp;
            fy = fp;
            dy = dp;
        else:
            if sgnd < 0.0:
                sty = stx
                fy = fx
                dy = dx
            stx = stp
            fx = fp
            dx = dp

        # Compute the new step and safeguard it.
        stpf = np.minimum(stpmax, stpf);
        stpf = np.maximum(stpmin, stpf);
        stp = stpf
        if brackt & bound:
            if sty > stx:
                stp = np.minimum(stx + self.delta * (sty - stx), stp)
            else:
                stp = np.maximum(stx + self.delta * (sty - stx), stp)

        return stx, fx, dx, sty, fy, dy, stp, fp, dp, brackt, success

    def run(self, problem, modl, dmodl, logger=None):
        """Method to apply CvSrch stepper"""
        # Writing to log file if any
        if logger:
            logger.addToLog("CVSRCH STEPPER BY STEP-LENGTH BRACKETING")
            logger.addToLog("xtol=%.2e ftol=%.2e gtol=%.2e alpha_min=%.2e alpha_max=%.2e maxfev=%d xtrapf=%.2e"
                            % (
                            self.xtol, self.ftol, self.gtol, self.alpha_min, self.alpha_max, self.maxfev, self.xtrapf))
        success = False
        # Obtain objective function for provided model
        phi_init = problem.get_obj(modl)
        # Getting pointer to problem's gradient vector
        prblm_grad = problem.get_grad(modl)
        dphi_init = prblm_grad.dot(dmodl)
        if dphi_init > 0.0:
            if logger:
                logger.addToLog("\tWarning! Current search direction is not a descent one!")
            return self.alpha, success
        # Model temporary vector
        model_step = modl.clone()
        # Getting pointer to problem's model vector
        prblm_mdl = problem.get_model()
        # Initial step length value
        alpha = deepcopy(self.alpha)
        # Estimating initial step length value
        if alpha < self.zero:
            alpha = self.estimate_initial_guess(problem, modl, dmodl, logger)
        if logger:
            logger.addToLog("\tinitial-steplength=%.2e" % alpha)

        # Initializing parameters
        p5 = 0.5
        cstep_success = True
        fev = 0
        width = self.alpha_max - self.alpha_min
        width1 = 2 * width
        brackt = False
        stage1 = True
        dphi_test = self.ftol * dphi_init

        # The variables alphax, phix, dphix contain the values of the step, function, and directional derivative at the best step.
        # The variables alphay, phiy, dphiy contain the value of the step, function, and derivative at the other endpoint of the interval of uncertainty.
        # The variables alpha, phi_c, dphi_c contain the values of the step, function, and derivative at the current step.
        alphax = 0.0
        phix = phi_init
        dphix = dphi_init
        alphay = 0.0
        phiy = phi_init
        dphiy = dphi_init

        # Start testing iteration
        while True:
            # Set the minimum and maximum steps to correspond to the present interval of uncertainty.
            if brackt:
                alpha_int_min = np.minimum(alphax, alphay)
                alpha_int_max = np.maximum(alphax, alphay)
            else:
                alpha_int_min = alphax
                alpha_int_max = alpha + self.xtrapf * (alpha - alphax)

            # Force the step to be within the bounds alpha_max and alpha_min.
            alpha = np.maximum(alpha, self.alpha_min)
            alpha = np.minimum(alpha, self.alpha_max)

            # If an unusual termination is to occur then choose alpha be the lowest point obtained so far.
            if ((brackt and (alpha <= alpha_int_min or alpha >= alpha_int_max)) or fev >= self.maxfev - 1 or (
                    not cstep_success) or (brackt and alpha_int_max - alpha_int_min <= self.xtol * alpha_int_max)):
                if logger:
                    logger.addToLog("\tUnusual termination is to occur. Setting alpha to be the lowest point obtained so far.")
                alpha = alphax

            # Evaluate the function and gradient at alpha and compute the directional derivative.
            if logger:
                logger.addToLog("\tCurrent testing point (alpha=%.2e): m_current+alpha*dm" % alpha)
            model_step.copy(modl)
            model_step.scaleAdd(dmodl, sc2=alpha)
            # Checking if model parameters hit the bounds
            problem.set_model(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if prblm_mdl.isDifferent(model_step):
                # Model hit bounds
                if logger:
                    logger.addToLog("\tModel hit provided bounds. Projecting it onto them.")
            phi_alpha = problem.get_obj(model_step)
            fev += 1
            # Checking if a NaN is encountered
            if isnan(phi_alpha):
                if logger:
                    logger.addToLog("\t!!!Objective function is NaN! Stepper unsuccessful!!!")
                problem.set_model(modl)
                break
            if logger:
                logger.addToLog("\tObjective function value of %.5e (feval = %d)" % (phi_alpha, problem.get_fevals()))
            prblm_grad = problem.get_grad(model_step)
            dphi_alpha = prblm_grad.dot(dmodl)
            phi_test1 = phi_init + alpha * dphi_test

            # Test for convergence
            if (brackt and (alpha <= alpha_int_min or alpha >= alpha_int_max)) or (not cstep_success):
                if logger:
                    logger.addToLog("\tRounding errors prevent further progress. There may not be a step which satisfies "
                                    "the sufficient decrease and curvature conditions. Tolerances may be too small.")
                break
            if alpha == self.alpha_max and phi_alpha <= phi_test1 and dphi_alpha <= dphi_test:
                if logger:
                    logger.addToLog("\tThe step-length value is at the upper bound (alpha_max) of %.2e" % self.alpha_max)
                break
            if alpha == self.alpha_min and (phi_alpha > phi_test1 or dphi_alpha >= dphi_test):
                if logger:
                    logger.addToLog("\tThe step-length value is at the lower bound (alpha_min) of %.2e" % self.alpha_min)
                break
            if fev >= self.maxfev:
                if logger:
                    logger.addToLog("\tNumber of objective function evaluation reached maxfev of %d" % self.maxfev)
                break
            if brackt and alpha_int_max - alpha_int_min <= self.xtol * alpha_int_max:
                if logger:
                    logger.addToLog("\tRelative width of the interval of uncertainty is at most xtol of %.2e" % self.xtol)
                break
            if phi_alpha <= phi_test1 and abs(dphi_test) <= self.gtol * (-dphi_init) and phi_alpha < phi_init:
                success = True
                if logger:
                    logger.addToLog("\tThe sufficient decrease condition and the directional derivative condition hold "
                                    "(i.e., Strong Wolfe conditions met).\n	Stepper successuful for step length value"
                                    "of %.2e and objective function of %.2e (feval = %d)"
                                    % (alpha, phi_alpha, problem.get_fevals()))
                break

            # In the first stage we seek a step for which the modified function has a nonpositive value and
            # nonnegative derivative.
            if stage1 and (phi_alpha <= phi_test1) and (dphi_alpha >= np.minimum(self.ftol, self.gtol) * dphi_init):
                stage1 = False

            # A modified function is used to predict the step only if
            # we have not obtained a step for which the modified
            # function has a nonpositive function value and nonnegative
            # derivative, and if a lower function value has been
            # obtained but the decrease is not sufficient.

            if stage1 and (phi_alpha <= phix) and (phi_alpha > phi_test1):
                # Define the modified function and derivative values.
                phim = phi_alpha - alpha * dphi_test
                phixm = phix - alphax * dphi_test
                phiym = phiy - alphay * dphi_test
                dphim = dphi_alpha - dphi_test
                dphixm = dphix - dphi_test
                dphiym = dphiy - dphi_test

                # Call cstep to update the interval of uncertainty and to compute the new step.
                [alphax, phixm, dphixm, alphay, phiym, dphiym, alpha, phim, dphim, brackt, cstep_success] = self.cstep(
                    alphax, phixm, dphixm, alphay, phiym, dphiym, alpha, phim, dphim, brackt, alpha_int_min,
                    alpha_int_max, logger)

                # Reset the function and gradient values for phi.
                phix = phixm + alphax * dphi_test
                phiy = phiym + alphay * dphi_test
                dphix = dphixm + dphi_test
                dphiy = dphiym + dphi_test

            else:
                # Call cstep to update the interval of uncertainty and to compute the new step.
                [alphax, phix, dphix, alphay, phiy, dphiy, alpha, phi_alpha, dphi_alpha, brackt,
                 cstep_success] = self.cstep(alphax, phix, dphix, alphay, phiy, dphiy, alpha, phi_alpha, dphi_alpha,
                                             brackt, alpha_int_min, alpha_int_max, logger)

            # Force a sufficient decrease in the size of the interval of uncertainty.
            if brackt:
                if abs(alphay - alphax) >= self.delta * width1:
                    alpha = alphax + p5 * (alphay - alphax)
                width1 = width
                width = abs(alphay - alphax)

            # End of iteration.

        if success:
            # Line search has finished, update model
            self.alpha = deepcopy(alpha)
            modl.copy(model_step)

        # Delete temporary vectors
        del model_step
        return alpha, success


class ParabolicStep(Stepper):
    """Parabolic Stepper class with three-point interpolation"""

    def __init__(self, c1=1.0, c2=2.0, ntry=10, alpha=0., alpha_scale_min=1.0e-10, alpha_scale_max=2000.00, shrink=0.25,
                 eval_parab=True):
        """
           Constructor for parabolic stepper with three-point interpolation:
           c1  		   	   = [1.0] - float; Scaling factor of first search point (i.e., m1 = c1*alpha*dm + m_current)
           c2  		   	   = [2.0] - float; Scaling factor of first search point (i.e., m2 = c2*alpha*dm + m_current)
           ntry  	   	   = [10] - integer; Number of trials for finding the step length
           alpha 		   = [0.] - float; Initial step-length guess
           alpha_scale_min = [1.0e-10] - float; Minimum scaling factor (c_optimal) for step-length allowed
           alpha_scale_max = [1000.00] - float; Maximum scaling factor (c_optimal) for step-length allowed
           shrink 		   = [0.25] - float; Shrinking factor if step length is not found at a given trial
           eval_parab 	   = [True] - boolean; Force parabola minimum to be computed. If False, the best point will be chosen from c1 or c2 and the parabola minimum is computed if necessary
        """
        self.c1 = c1  # Scaling for first tested point
        self.c2 = c2  # Scaling for second tested point
        self.ntry = ntry  # Number of total trials before re-estimating initial alpha value
        self.alpha = alpha  # Initial step length guess
        self.alpha_scale_min = alpha_scale_min  # Maximum scaling value for the step length
        self.alpha_scale_max = alpha_scale_max  # Minimum scaling value for the step length
        self.shrink = shrink  # Shrinking scaling factor if trial is unsuccessful
        self.zero = 10 ** (np.floor(
            np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow
        self.eval_parab = eval_parab
        return

    def run(self, problem, modl, dmodl, logger=None):
        """Method to apply parabolic stepper"""
        # Writing to log file if any
        global obj1
        if logger:
            logger.addToLog("PARABOLIC STEPPER USING THREE-POINT INTERPOLATION")
            logger.addToLog("c1=%.2e c2=%.2e ntry=%d steplength-scaling-min=%.2e steplength-scaling-max=%.2e shrinking-factor=%.2e"
                            % (self.c1, self.c2, self.ntry, self.alpha_scale_min, self.alpha_scale_max, self.shrink))
        success = False
        # Obtain objective function for provided model
        obj0 = problem.get_obj(modl)
        # Model temporary vector
        model_step = modl.clone()
        # Getting pointer to problem's model vector
        prblm_mdl = problem.get_model()
        # Initial step length value
        alpha = deepcopy(self.alpha)
        # Checking if current search direction is a descending one
        prblm_grad = problem.get_grad(prblm_mdl)
        dphi = prblm_grad.dot(dmodl)
        if dphi > 0.0:
            if logger:
                logger.addToLog("\tWarning! Current search direction is not a descent one!")
            return alpha, success
        itry = 1
        total_trials = deepcopy(self.ntry)
        if alpha != 0.:
            # If initial step length is different than zero, we tests twice in case we need to re-estimate initial alpha
            total_trials *= 2
        while itry <= total_trials:
            # Writing info to log file
            if logger:
                logger.addToLog("\ttrial number: %d" % itry)
                logger.addToLog("\tinitial-steplength=%.2e" % alpha)
            # Find the first guess as if the problem was linear (Tangent method)
            if (itry == self.ntry) or (alpha < self.zero):
                alpha = self.estimate_initial_guess(problem, modl, dmodl, logger)
                if logger:
                    logger.addToLog("\tGuessing step length of: %.2e" % alpha)
            # Test values of objective function for two scaled versions of the step length
            # Testing c1 scale
            if logger:
                logger.addToLog("\tTesting point (c1=%.2e): m_current+c1*alpha*dm" % self.c1)
            model_step.copy(modl)
            model_step.scaleAdd(dmodl, sc2=self.c1 * alpha)
            # Checking if model parameters hit the bounds
            problem.set_model(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if prblm_mdl.isDifferent(model_step):
                # Model hit bounds
                msg = "\tModel hit provided bounds. Projecting it onto them."
                if logger:
                    logger.addToLog(msg)
            obj1 = problem.get_obj(model_step)
            # Copying residuals for point c1
            res_prblm = problem.get_res(model_step)
            res1 = res_prblm.clone()
            if logger:
                logger.addToLog("\tObjective function value of %.5e" % obj1)
            # Checking if a NaN is encountered in any of the two tested points
            if isnan(obj1):
                if logger:
                    logger.addToLog("\t!!!Problem with step length and objective function!!!")
                if itry >= self.ntry:
                    if logger:
                        logger.addToLog("\t!!!Check problem definition or change solver!!!")
                    # Setting model to current one and resetting initial step length value
                    alpha = 0.0
                    self.alpha = 0.0
                    problem.set_model(modl)
                    break
                else:
                    if logger:
                        logger.addToLog("\t!!!Guessing linear step length to try to solve problem!!!")
                    itry = self.ntry  # To not repeat computation of linear guess
                    continue
            # Testing c2 scale
            msg = "\tTesting point (c2=%.2e): m_current+c2*alpha*dm" % self.c2
            if logger:
                logger.addToLog(msg)
            model_step.copy(modl)
            model_step.scaleAdd(dmodl, sc2=self.c2 * alpha)
            # Checking if model parameters hit the bounds
            problem.set_model(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if prblm_mdl.isDifferent(model_step):
                # Model hit bounds
                msg = "\tModel hit provided bounds. Projecting it onto them."
                if logger:
                    logger.addToLog(msg)
            obj2 = problem.get_obj(model_step)
            # Copying residuals for point c1
            res_prblm = problem.get_res(model_step)
            res2 = res_prblm.clone()
            if logger:
                logger.addToLog("\tObjective function value of %.5e" % obj2)
            # Checking for NaN
            if isnan(obj2):
                if logger:
                    logger.addToLog("\t!!!Problem with step length and objective function!!!")
                if itry >= self.ntry:
                    if logger:
                        logger.addToLog("\t!!!Check problem definition or change solver!!!")
                    # Setting model to current one and resetting initial step length value
                    alpha = 0.0
                    self.alpha = 0.0
                    problem.set_model(modl)
                    break
                else:
                    if logger:
                        logger.addToLog("\t!!!Guessing linear step length to try to solve problem!!!")
                    itry = self.ntry  # To not repeat computation of linear guess
                    continue
            # Checking if parabolic point is necessary or not
            if not self.eval_parab:
                # Setting third point to infinity
                obj3 = np.inf
                # Check which one is the best step length
                msg = "\n\tAs requested, parabola minimum was not evaluated! Unless necessary!"
                if obj1 < obj0 and obj1 < obj2 and obj1 < obj3:
                    success = True
                    alpha *= self.c1
                    if logger:
                        logger.addToLog("\tc1 best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals() - 1) + msg)
                    break
                elif obj2 < obj0 and obj2 < obj1 and obj2 < obj3:
                    success = True
                    alpha *= self.c2
                    if logger:
                        logger.addToLog("\tc2 best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals()) + msg)
                    break
            # If points lay on a horizontal line pick minimum alpha set by user
            if obj0 == obj1 == obj2 or (self.c2 * (obj1 - obj0) + self.c1 * (obj0 - obj2)) == 0.:
                step_scale = self.alpha_scale_min
                if logger:
                    logger.addToLog("\tTwo testing points on a line: cannot fit a parabola, using minimum step-length of %.2e"
                                    % (step_scale * alpha))
            else:
                # Otherwise, find the optimal parabolic step length
                step_scale = 0.5 * (self.c2 * self.c2 * (obj1 - obj0) + self.c1 * self.c1 * (obj0 - obj2)) / (
                        self.c2 * (obj1 - obj0) + self.c1 * (obj0 - obj2))
                if logger:
                    logger.addToLog("\tTesting point (c_opt=%.2e): m_current+c_opt*alpha*dm (parabola minimum)" % step_scale)
            # If step length negative, re-evaluate points
            if step_scale < 0.:
                if logger:
                    logger.addToLog("\tEncountered a negative step-length value: %.2e; Setting parabola-minimum objective function to infinity."
                                    % (step_scale * alpha))
                # Skipping parabola minimum and setting obj3 to infinity
                obj3 = np.inf
            else:
                # Clipping the step-length scale
                if step_scale < self.alpha_scale_min:
                    if logger:
                        logger.addToLog("\t!!! step-length scale of %.2e smaller than provided lower bound."
                                        "Clipping its value to bound value of %.2e !!!" % (step_scale, self.alpha_scale_min))
                    step_scale = self.alpha_scale_min
                elif step_scale > self.alpha_scale_max:
                    if logger:
                        logger.addToLog("\t!!! step-length scale of %.2e greater than provided upper bound."
                                        "Clipping its value to bound value of %.2e !!!" % (step_scale, self.alpha_scale_max))
                    step_scale = self.alpha_scale_max

                # Testing parabolic scale
                # Compute new objective function at the minimum of the parabolic approximation
                model_step.copy(modl)
                model_step.scaleAdd(dmodl, sc2=step_scale * alpha)
                # Checking if model parameters hit the bounds
                problem.set_model(model_step)
                # Projecting model onto the bounds (if any)
                if "bounds" in dir(problem):
                    problem.bounds.apply(model_step)
                if prblm_mdl.isDifferent(model_step):
                    # Model hit bounds
                    msg = "\tModel hit provided bounds. Projecting it onto them."
                    if logger:
                        logger.addToLog(msg)
                obj3 = problem.get_obj(model_step)
                if logger:
                    logger.addToLog("\tObjective function value of %.5e" % obj3)

            # Writing info to log file
            if logger:
                logger.addToLog("\tInitial objective function value: %.5e,"
                                "Objective function at c1*alpha*dm: %.5e,"
                                "Objective function at c2*alpha*dm: %.5e,"
                                "Objective function at parabola minimum: %.5e"
                                % (obj0, obj1, obj2, obj3))
            itry += 1

            # Check which one is the best step length
            if obj1 < obj0 and obj1 < obj2 and obj1 < obj3:
                success = True
                alpha *= self.c1
                if logger:
                    logger.addToLog("\tc1 best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals() - 2))
                break
            elif obj2 < obj0 and obj2 < obj1 and obj2 < obj3:
                success = True
                alpha *= self.c2
                if logger:
                    logger.addToLog("\tc2 best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals() - 1))
                break
            elif obj3 < obj0 and obj3 <= obj1 and obj3 <= obj2:
                success = True
                alpha *= step_scale
                if logger:
                    logger.addToLog("\tparabola minimum best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals()))
                break
            else:
                # Shrink line search
                alpha *= self.shrink
                if logger:
                    logger.addToLog("\tShrinking search direction")

        if success:
            # Line search has finished, update model
            self.alpha = deepcopy(alpha)
            model_step.copy(modl)  # model_step = m_current
            model_step.scaleAdd(dmodl, sc2=self.alpha)
            # Checking if model parameters hit the bounds
            modl.copy(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if modl.isDifferent(model_step):
                # Computing true scaled search direction dm = m_new_clipped - m_current
                dmodl.copy(model_step)
                dmodl.scaleAdd(modl, 1.0, -1.0)
                # Scaled by the inverse of the step length
                dmodl.scale(1.0 / self.alpha)
            # Setting model and residual vectors to c1 or c2 point if parabola minimum is not picked
            problem.set_model(model_step)
            if obj1 < obj0 and obj1 < obj2 and obj1 < obj3:
                problem.set_residual(res1)
            elif obj2 < obj0 and obj2 < obj1 and obj2 < obj3:
                problem.set_residual(res2)
            modl.copy(model_step)
        # Delete temporary vectors
        del model_step, res1, res2
        return alpha, success


class ParabolicStepConst(Stepper):
    """Parabolic Stepper class assuming constant local curvature"""

    def __init__(self, c1=1.0, ntry=10, alpha=0., alpha_scale_min=1.0e-10, alpha_scale_max=2000.00, shrink=0.25):
        """
           Constructor for parabolic stepper assuming constant local curvature:
           c1  		   	   = [1.0] - float; Scaling factor of the search point (i.e., m1 = c1*alpha*dm + m_current)
           ntry  	   	   = [10] - integer; Number of trials for finding the step length
           alpha 		   = [0.] - float; Initial step-length guess
           alpha_scale_min = [1.0e-10] - float; Minimum scaling factor (c_optimal) for step-length allowed
           alpha_scale_max = [1000.00] - float; Maximum scaling factor (c_optimal) for step-length allowed
           shrink 		   = [0.25] - float; Shrinking factor if step length is not found at a given trial
        """
        self.c1 = c1  # Scaling for first tested point
        self.ntry = ntry  # Number of total trials before re-estimating initial alpha value
        self.alpha = alpha  # Initial step length guess
        self.alpha_scale_min = alpha_scale_min  # Minimum scaling value for the step length
        self.alpha_scale_max = alpha_scale_max  # Maximum scaling value for the step length
        self.shrink = shrink  # Shrinking scaling factor if trial is unsuccessful
        self.zero = 10 ** (np.floor(
            np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow
        return

    def run(self, problem, modl, dmodl, logger=None):
        """Method to apply parabolic stepper"""
        # Writing to log file if any
        if logger:
            logger.addToLog("PARABOLIC STEPPER ASSUMING CONSTANT LOCAL CURVATURE")
            logger.addToLog("c1=%.2e ntry=%d steplength-scaling-min=%.2e steplength-scaling-max=%.2e shrinking-factor=%.2e"
                            % (self.c1, self.ntry, self.alpha_scale_min, self.alpha_scale_max, self.shrink))
        success = False
        # Obtain objective function for provided model
        obj0 = problem.get_obj(modl)
        # Model temporary vector
        model_step = modl.clone()
        # Getting pointer to problem's model vector
        prblm_mdl = problem.get_model()
        # Initial step length value
        alpha = deepcopy(self.alpha)
        # Getting pointer to problem's gradient vector
        prblm_grad = problem.get_grad(prblm_mdl)
        dphi = prblm_grad.dot(dmodl)
        if dphi > 0.0:
            if logger:
                logger.addToLog("\tWarning! Current search direction is not a descent one!")
            return alpha, success
        itry = 1
        total_trials = deepcopy(self.ntry)
        if alpha != 0.:
            # If initial step length is different than zero, we tests twice in case we need to re-estimate initial alpha
            total_trials *= 2
        while itry <= total_trials:
            # Writing info to log file
            if logger:
                logger.addToLog("\ttrial number: %d" % itry)
                logger.addToLog("\tinitial-steplength=%.2e" % alpha)
            # Find the first guess as if the problem was linear (Tangent method)
            if (itry == self.ntry) or (alpha < self.zero):
                alpha = self.estimate_initial_guess(problem, modl, dmodl, logger)
                if logger:
                    logger.addToLog("\tGuessing step length of: %.2e" % alpha)
            # Test values of objective function for two scaled versions of the step length
            # Testing c1 scale
            if logger:
                logger.addToLog("\tTesting point (c1=%.2e): m_current+c1*alpha*dm" % self.c1)
            model_step.copy(modl)
            model_step.scaleAdd(dmodl, sc2=self.c1 * alpha)
            # Checking if model parameters hit the bounds
            problem.set_model(model_step)
            # Projecting model onto the bounds (if any) and rotate search direction
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if prblm_mdl.isDifferent(model_step):
                # Model hit bounds
                msg = "\tModel hit provided bounds. Projecting it onto them."
                if logger:
                    logger.addToLog(msg)
                # Computing true scaled search direction dm = m_new_clipped - m_current
                dmodl.copy(model_step)
                dmodl.scaleAdd(modl, 1.0, -1.0)
                # Scaled by the inverse of the step length
                dmodl.scale(1.0 / (self.c1 * alpha))
            obj1 = problem.get_obj(model_step)
            # Copying residuals for point c1
            res_prblm = problem.get_res(model_step)
            res1 = res_prblm.clone()
            if logger:
                logger.addToLog("\tObjective function value of %.5e" % obj1)
            # Checking if a NaN is encountered in any of the two tested points
            if isnan(obj1):
                if logger:
                    logger.addToLog("\t!!!Problem with step length and objective function!!!")
                if itry >= self.ntry:
                    if logger:
                        logger.addToLog("\t!!!Check problem definition or change solver!!!")
                    # Setting model to current one and resetting initial step length value
                    alpha = 0.0
                    self.alpha = 0.0
                    problem.set_model(modl)
                    break
                else:
                    if logger:
                        logger.addToLog("\t!!!Guessing linear step length to try to solve problem!!!")
                    itry = self.ntry  # To not repeat computation of linear guess
                    continue
            # Computing local constant curvature
            phi_der = prblm_grad.dot(dmodl)  # First derivative of the objective function with respect to alpha
            c = 2.0 * ((obj1 - obj0) / (self.c1 * alpha * self.c1 * alpha) - phi_der / (self.c1 * alpha))
            # Checking the curvature value
            if c <= 0.:
                # Shrink line search
                alpha *= self.shrink
                if logger:
                    logger.addToLog("\tEstimated a negative curvature of %.2e. Shrinking search direction" % c)
                itry += 1
                continue
            # Computing objective function at local parabola minimum
            alpha_parab = - phi_der / c
            step_scale = alpha_parab / alpha
            if logger:
                logger.addToLog("\tTesting point (c_opt=%.2e): m_current+c_opt*alpha*dm (parabola minimum)" % step_scale)
            # If step length negative, re-evaluate points
            if alpha_parab < 0.:
                if logger:
                    logger.addToLog("\tEncountered a negative step-length value: %.2e; Shrinking step-length value." % alpha_parab)
                # Shrink line search
                alpha *= self.shrink
                itry += 1
                continue
            # Clipping the step-length scale
            if step_scale < self.alpha_scale_min:
                if logger:
                    logger.addToLog("\t!!! step-length scale of %.2e smaller than provided lower bound."
                                    "Clipping its value to bound value of %.2e !!!" % (step_scale, self.alpha_scale_min))
                step_scale = self.alpha_scale_min
            elif step_scale > self.alpha_scale_max:
                if logger:
                    logger.addToLog("\t!!! step-length scale of %.2e greater than provided upper bound."
                                    "Clipping its value to bound value of %.2e !!!" % (step_scale, self.alpha_scale_max))
                step_scale = self.alpha_scale_max

            # Testing parabolic scale
            # Compute new objective function at the minimum of the parabolic approximation
            model_step.copy(modl)
            model_step.scaleAdd(dmodl, sc2=alpha * step_scale)
            # Checking if model parameters hit the bounds
            problem.set_model(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if prblm_mdl.isDifferent(model_step):
                # Model hit bounds
                if logger:
                    logger.addToLog("\tModel hit provided bounds. Projecting it onto them.")
            obj2 = problem.get_obj(model_step)
            if logger:
                logger.addToLog("\tObjective function value of %.5e" % obj2)

            # Writing info to log file
            if logger:
                logger.addToLog("\tInitial objective function value: %2e,"
                                "Objective function at c1*alpha*dm: %.2e,"
                                "Objective function at parabola minimum: %.2e"
                                % (obj0, obj1, obj2))
            itry += 1

            # Check which one is the best step length
            if obj1 < obj0 and obj1 < obj2:
                success = True
                alpha *= self.c1
                if logger:
                    logger.addToLog("\tc1 best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals() - 1))
                break
            elif obj2 < obj0 and obj2 <= obj1:
                success = True
                alpha *= step_scale
                if logger:
                    logger.addToLog("\tparabola minimum best step-length value of: %.2e (feval = %d)" % (alpha, problem.get_fevals()))
                break
            else:
                # Shrink line search
                alpha *= self.shrink
                if logger:
                    logger.addToLog("\tShrinking search direction")

        if success:
            # Line search has finished, update model
            self.alpha = deepcopy(alpha)
            model_step.copy(modl)  # model_step = m_current
            model_step.scaleAdd(dmodl, sc2=self.alpha)
            # Checking if model parameters hit the bounds
            modl.copy(model_step)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(model_step)
            if modl.isDifferent(model_step):
                # Computing true scaled search direction dm = m_new_clipped - m_current
                dmodl.copy(model_step)
                dmodl.scaleAdd(modl, 1.0, -1.0)
                # Scaled by the inverse of the step length
                dmodl.scale(1.0 / self.alpha)
            # Setting model and residual vectors to c1 or c2 point if parabola minimum is not picked
            problem.set_model(model_step)
            if obj1 < obj0 and obj1 < obj2:
                problem.set_residual(res1)
            modl.copy(model_step)
        # Delete temporary vectors
        del model_step, res1
        return alpha, success

class StrongWolfe(Stepper):
    """Compute a line search to satisfy the strong Wolfe conditions.
       Algorithm 3.5. Page 60. "Numerical Optimization". Nocedal & Wright.
       Implementation based on the ones in the GitHub repo: https://github.com/bgranzow/L-BFGS-B.git
       """

    def __init__(self, c1=1.e-4, c2=0.9 , ntry=20, alpha=1., alpha_scale=0.8, alpha_max=2.5, keepAlpha=False):
        """
           Constructor for parabolic stepper assuming constant local curvature:
           c1  		   	   = [1.e-4] - float; c1 value to tests first Wolfe condition (should be between 0 and 1)
           c2  		   	   = [0.9] - float; c2 value to tests second Wolfe condition (should be between c1 and 1). For Quasi-Newton (e.g., L-BFGS) choose default. Otherwise, for other methods (e.g., NLCG) choose 0.1
           ntry  	   	   = [20] - integer; Number of trials for finding the step length
           alpha 		   = [1.] - float; Initial step-length guess
           alpha_scale 	   = [0.8] - float; step-length update factor used for updating step length guess
           alpha_max       = [2.5] - float; Maximum step-length value allowed
           keepAlpha       = [False] - boolean; Whether to keep or forget previously found step-length value
        """
        self.c1 = c1
        self.c2 = c2
        self.ntry = ntry  # Number of total trials before re-estimating initial alpha value
        self.alpha = alpha  # Initial step length guess
        self.alpha_max = alpha_max  # Maximum step-length value
        self.alpha_scale = alpha_scale
        self.zero = 10 ** (np.floor(
            np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow
        self.keepAlpha = keepAlpha
        return

    def alpha_zoom(self, problem, mdl0, mdl, obj0, dphi0, dmodl, alpha_lo, alpha_hi, logger=None):
        """Algorithm 3.6, Page 61. "Numerical Optimization". Nocedal & Wright."""
        itry = 0
        alpha = 0.0
        while itry < self.ntry:
            if logger:
                logger.addToLog("\t\ttrial number [alpha_zoom]: %d" % (itry+1))
            alpha_i = 0.5 * (alpha_lo + alpha_hi)
            alpha = alpha_i
            # x = x0 + alpha_i * p
            mdl.copy(mdl0)
            mdl.scaleAdd(dmodl, sc2=alpha_i)
            # Evaluating objective and gradient function
            obj_i = problem.get_obj(mdl)
            if logger:
                logger.addToLog("\t\tObjective function value of %.5e at m_i with alpha=%.5e [alpha_zoom]" %(obj_i, alpha_i))
            if isnan(obj_i):
                if logger:
                    logger.addToLog("\t\t!!!Problem with step length and objective function; Setting alpha = 0.0 [alpha_zoom]!!!")
                alpha = 0.0
                break
            grad_i = problem.get_grad(mdl)
            # x_lo = x0 + alpha_lo * p;
            mdl.copy(mdl0)
            mdl.scaleAdd(dmodl, sc2=alpha_lo)  # x = x0 + alpha_i * p;
            obj_lo = problem.get_obj(mdl)
            if logger:
                logger.addToLog("\t\tObjective function value of %.5e at m_lo with alpha_lo=%.5e [alpha_zoom]" %(obj_lo, alpha_lo))
            if isnan(obj_lo):
                if logger:
                    logger.addToLog("\t\t!!!Problem with step length and objective function; Setting alpha = 0.0 [alpha_zoom]!!!")
                alpha = 0.0
                break
            if (obj_i > obj0 + self.c1 * alpha_i * dphi0) or (obj_i >= obj_lo):
                alpha_hi = alpha_i
            else:
                dphi = grad_i.dot(dmodl)
                if np.abs(dphi) <= -self.c2 * dphi0:
                    alpha = alpha_i
                    break
                if dphi * (alpha_hi - alpha_lo) >= 0.0:
                    alpha_hi = alpha_lo
                alpha_lo = alpha_i
            # Increasing trial counter of alpha zoom
            itry += 1
            if itry > self.ntry:
                alpha = alpha_i
                break
        return alpha

    def run(self, problem, modl, dmodl, logger=None):
        """Method to apply line search that satisfies strong Wolfe conditions"""
        # Writing to log file if any
        if logger:
            logger.addToLog("STRONG-WOLFE STEPPER")
            logger.addToLog(
                "c1=%.2e c2=%.2e ntry=%d alpha-max=%.2e keepAlpha=%s"
                % (self.c1, self.c2, self.ntry, self.alpha_max, self.keepAlpha))
        success = False
        # Obtain objective function for provided model
        obj0 = problem.get_obj(modl)
        obj_im1 = deepcopy(obj0)
        # Model temporary vector
        model_step = modl.clone()
        # Initial step length value
        alpha_i = deepcopy(self.alpha)
        alpha_im1 = 0.0
        alpha = 0.0
        # Getting pointer to problem's gradient vector
        prblm_grad = problem.get_grad(modl).clone()
        dphi0 = prblm_grad.dot(dmodl)
        itry = 0
        while itry < self.ntry:
            # Writing info to log file
            if logger:
                logger.addToLog("\ttrial number: %d" % (itry+1))
                logger.addToLog("\tinitial-steplength=%.2e" % alpha_i)
            # Find the first guess as if the problem was linear (Tangent method)
            if alpha_i <= self.zero:
                alpha_i = self.estimate_initial_guess(problem, modl, dmodl, logger)
                self.alpha_max *= alpha_i
                if logger:
                    logger.addToLog("\tGuessing step length of: %.2e" % alpha_i)

            # Updating model point
            model_step.copy(modl) # x = x0
            model_step.scaleAdd(dmodl, sc2=alpha_i) # x = x0 + alpha_i * p;
            # Evaluating objective and gradient function
            obj_i = problem.get_obj(model_step)
            if logger:
                logger.addToLog("\tObjective function value of %.5e" % obj_i)
            if isnan(obj_i):
                if logger:
                    logger.addToLog("\t!!!Problem with step length and objective function; Stopping stepper!!!")
                break
            grad_i = problem.get_grad(model_step)
            if (obj_i > obj0 + self.c1 * dphi0) or ((itry > 1) and (obj_i >= obj_im1)):
                alpha = self.alpha_zoom(problem, modl, model_step, obj0, dphi0, dmodl, alpha_im1, alpha_i, logger)
                if logger:
                    logger.addToLog("\tCondition 1 matched; step-length value of: %.2e" % alpha)
                success = True
                break

            # dphi = transpose(g_i) * p;
            dphi = grad_i.dot(dmodl)
            if np.abs(dphi) <= -self.c2 * dphi0:
                alpha = alpha_i
                if logger:
                    logger.addToLog("\tCondition 2 matched; step-length value of: %.2e" % alpha)
                success = True
                break
            if dphi >= self.zero:
                alpha = self.alpha_zoom(problem, modl, model_step, obj0, dphi0, dmodl, alpha_i, alpha_im1, logger)
                if logger:
                    logger.addToLog("\tCondition 3 matched; step-length value of: %.2e" % alpha)
                success = True
                break

            # Update step-length value
            alpha_im1 = alpha_i
            obj_im1 = obj_i
            alpha_i = alpha_i + self.alpha_scale * (self.alpha_max - alpha_i)

            if itry > self.ntry:
                alpha = alpha_i
                if logger:
                    logger.addToLog("\tMaximum number of trials reached (ntry = %d); step-length value of: %.2e" %(self.ntry,alpha))
                success = True
                break

            # Update trial number
            itry += 1
        if success:
            # Line search has finished, update model
            modl.scaleAdd(dmodl, sc2=alpha)
            if self.keepAlpha:
                self.alpha = alpha
        # Delete temporary vectors
        del model_step

        return alpha, success


from .matrix import *
from .signal import *
from .derivative import *

__all__ = [
    "Matrix",
    "FirstDerivative",
    "SecondDerivative",
    "Laplacian",
    "Gradient",
    "ConvND",
    "GaussianFilter",
    "ZeroPad",
]



import os
import numpy as np
import h5py

from occamypy.utils import sep


class Vector:
    """Abstract python vector class"""
    
    def __init__(self):
        """Default constructor"""
        self.shape = None
        self.size = None
        self.ndim = None
        self.type = type(self)
    
    def __repr__(self):
        return self.getNdArray().__repr__()
    
    def __del__(self):
        """Default destructor"""
    
    def __add__(self, other):  # self + other
        if type(other) in [int, float]:
            self.addbias(other)
            return self
        elif isinstance(other, Vector):
            self.scaleAdd(other)
            return self
        else:
            raise TypeError('Argument has to be either scalar or vector, got %r instead' % other)
    
    def __sub__(self, other):  # self - other
        self.__add__(-other)
        return self
    
    def __neg__(self):  # -self
        self.scale(-1)
        return self
    
    def __mul__(self, other):  # self * other
        if type(other) in [int, float]:
            self.scale(other)
            return self
        elif isinstance(other, Vector):
            self.multiply(other)
            return self
        else:
            raise NotImplementedError
    
    def __rmul__(self, other):
        if type(other) in [int, float]:
            self.scale(other)
            return self
        elif isinstance(other, Vector):
            self.multiply(other)
            return self
        else:
            raise NotImplementedError
    
    def __pow__(self, power, modulo=None):
        if type(power) in [int, float]:
            self.pow(power)
        else:
            raise TypeError('power has to be a scalar')
    
    def __abs__(self):
        self.abs()
    
    def __truediv__(self, other):  # self / other
        if type(other) in [int, float]:
            self.scale(1 / other)
        elif isinstance(other, Vector):
            self.multiply(other.clone().reciprocal())
        else:
            raise TypeError('other has to be either a scalar or a vector')
    
    def __getitem__(self, item):
        return self.getNdArray()[item]
    
    def __setitem__(self, key, value):
        self.getNdArray()[key] = value
    
    @property
    def whoami(self):
        return type(self).__name__
    
    # Class vector operations
    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        raise NotImplementedError("getNdArray must be overwritten")
    
    def norm(self, N=2):
        """Function to compute vector N-norm"""
        raise NotImplementedError("norm must be overwritten")
    
    def zero(self):
        """Function to zero out a vector"""
        self.set(0)
        return self
    
    def max(self):
        """Function to obtain maximum value within a vector"""
        raise NotImplementedError("max must be overwritten")
    
    def min(self):
        """Function to obtain minimum value within a vector"""
        raise NotImplementedError("min must be overwritten")
    
    def set(self, val):
        """Function to set all values in the vector"""
        raise NotImplementedError("set must be overwritten")
    
    def scale(self, sc):
        """Function to scale a vector"""
        self.getNdArray()[:] *= sc
        return self
    
    def addbias(self, bias):
        """Function to add bias to a vector"""
        self.getNdArray()[:] += bias
        return self
    
    def rand(self):
        """Function to randomize a vector"""
        raise NotImplementedError("rand must be overwritten")
    
    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        raise NotImplementedError("clone must be overwritten")
    
    def cloneSpace(self):
        """Function to clone vector space"""
        raise NotImplementedError("cloneSpace must be overwritten")
    
    def checkSame(self, other):
        """Function to check to make sure the vectors exist in the same space"""
        return self.shape == other.shape
    
    def writeVec(self, filename, mode='w'):
        """Function to write vector to file"""
        # Check writing mode
        if mode not in 'wa':
            raise ValueError("Mode must be appending 'a' or writing 'w' ")
        # Construct ax_info if the object has getHyper
        if hasattr(self, "getHyper"):
            hyper = self.getHyper()
            self.ax_info = []
            for iaxis in range(hyper.getNdim()):
                self.ax_info.append([hyper.getAxis(iaxis + 1).n,
                                     hyper.getAxis(iaxis + 1).o,
                                     hyper.getAxis(iaxis + 1).d,
                                     hyper.getAxis(iaxis + 1).label])
        
        # check output file type
        _, ext = os.path.splitext(filename)  # file extension
        
        # SEP vector with header in filename.H and binary in DATAPATH/filename.H@
        if ext == ".H":
            # writing header/pointer file if not present and not append mode
            if not (os.path.isfile(filename) and mode in 'a'):
                binfile = sep.datapath + filename.split('/')[-1] + '@'
                with open(filename, mode) as f:
                    # Writing axis info
                    if self.ax_info:
                        for ii, ax_info in enumerate(self.ax_info):
                            ax_id = ii + 1
                            f.write("n%s=%s o%s=%s d%s=%s label%s='%s'\n" % (
                                ax_id, ax_info[0], ax_id, ax_info[1], ax_id, ax_info[2], ax_id, ax_info[3]))
                    else:
                        for ii, n_axis in enumerate(tuple(reversed(self.shape))):
                            ax_id = ii + 1
                            f.write("n%s=%s o%s=0.0 d%s=1.0 \n" % (ax_id, n_axis, ax_id, ax_id))
                    # Writing last axis for allowing appending (unless we are dealing with a scalar)
                    if self.shape != (1,):
                        ax_id = self.ndim + 1
                        f.write("n%s=%s o%s=0.0 d%s=1.0 \n" % (ax_id, 1, ax_id, ax_id))
                    f.write("in='%s'\n" % binfile)
                    esize = "esize=4\n"
                    if self.getNdArray().dtype == np.complex64:
                        esize = "esize=8\n"
                    f.write(esize)
                    f.write("data_format=\"native_float\"\n")
                f.close()
            else:
                binfile = sep.get_binary(filename)
                if mode in 'a':
                    axes = sep.get_axes(filename)
                    # Number of vectors already present in the file
                    if self.shape == (1,):
                        n_vec = axes[0][0]
                        append_dim = self.ndim
                    else:
                        n_vec = axes[self.ndim][0]
                        append_dim = self.ndim + 1
                    with open(filename, mode) as f:
                        f.write("n%s=%s o%s=0.0 d%s=1.0 \n" % (append_dim, n_vec + 1, append_dim, append_dim))
                    f.close()
            # Writing binary file
            fmt = '>f'
            if self.getNdArray().dtype == np.complex64 or self.getNdArray().dtype == np.complex128:
                format = '>c8'
            with open(binfile, mode + 'b') as f:
                # Writing big-ending floating point number
                if np.isfortran(self.getNdArray()):  # Forcing column-wise binary writing
                    # self.getNdArray().flatten('F').astype(fmt,copy=False).tofile(fid)
                    self.getNdArray().flatten('F').tofile(f, format=fmt)
                else:
                    # self.getNdArray().astype(fmt,order='C',subok=False,copy=False).tofile(fid)
                    self.getNdArray().tofile(f, format=fmt)
            f.close()
        
        # numpy dictionary
        elif ext == '.npy':
            if mode not in 'a':
                if self.ax_info:  # TODO fix saving of ax_info
                    axes = dict()
                    for ii, ax_info in enumerate(self.ax_info):
                        axes['%s' % ii + 1] = dict(n=ax_info[0], o=ax_info[1], d=ax_info[2], label=ax_info[3])
                    
                    np.save(file=filename,
                            arr=dict(arr=self.getNdArray(),
                                     ax_info=axes),
                            allow_pickle=True)
                else:
                    np.save(file=filename, arr=self.getNdArray(), allow_pickle=False)
        
        elif ext == '.h5':  # TODO implement saving to hdf5
            # https://moonbooks.org/Articles/How-to-save-a-large-dataset-in-a-hdf5-file-using-python--Quick-Guide/
            if mode not in 'a':
                with h5py.File(filename, 'wb') as f:
                    dset = f.create_dataset("vec", data=self.getNdArray())
            else:
                raise NotImplementedError
        
        else:
            raise ValueError("ERROR! Output format has to be H, npy, or h5")
        
        return
    
    # TODO implement on seplib
    def abs(self):
        """Return a vector containing the absolute values"""
        raise NotImplementedError('abs method must be implemented')
    
    # TODO implement on seplib
    def sign(self):
        """Return a vector containing the signs"""
        raise NotImplementedError('sign method have to be implemented')
    
    # TODO implement on seplib
    def reciprocal(self):
        """Return a vector containing the reciprocals of self"""
        self.getNdArray()[:] = 1. / self.getNdArray()
        return self
    
    # TODO implement on seplib
    def maximum(self, other):
        """Return a new vector of element-wise maximum of self and other"""
        raise NotImplementedError('maximum method must be implemented')
    
    # TODO implement on seplib
    def conj(self):
        """Compute the complex conjugate of the vector"""
        raise NotImplementedError('conj method must be implemented')
    
    def transpose(self):
        """Compute the transpose of the vector"""
        
    def hermitian(self):
        """Compute the hermitian, i.e. the conjugate transpose"""
        return self.transpose().conj()
    
    # TODO implement on seplib
    def pow(self, power):
        """Compute element-wise power of the vector"""
        raise NotImplementedError('pow method must be implemented')
    
    # TODO implement on seplib
    def real(self):
        """Return the real part of the vector"""
        raise NotImplementedError('real method must be implemented')
    
    # TODO implement on seplib
    def imag(self):
        """Return the imaginary part of the vector"""
        raise NotImplementedError('imag method must be implemented')
    
    # Combination of different vectors
    
    def copy(self, other):
        """Function to copy vector"""
        """Function to copy vector from input vector"""
        # Checking whether the input is a vector or not
        if not isinstance(other, self.type):
            raise TypeError("Provided input vector has to be of the same type!")
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError('Dimensionality not equal: self = %s; other = %s' % (self.shape, other.shape))
        # Element-wise copy of the input array
        self.getNdArray()[:] = other.getNdArray()
        return self
    
    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale two vectors and add them to the first one"""
        # Checking whether the input is a vector or not
        if not isinstance(other, self.type):
            raise TypeError("Provided input vector has to be of the same type!")
        # Checking dimensionality
        if not self.checkSame(other):
            raise ValueError("Dimensionality not equal: self = %s; other = %s" % (self.shape, other.shape))
        # Performing scaling and addition
        self.getNdArray()[:] = sc1 * self.getNdArray() + sc2 * other.getNdArray()
        return self
    
    def dot(self, other):
        """Function to compute dot product between two vectors"""
        raise NotImplementedError("dot must be overwritten")
    
    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        raise NotImplementedError("multiply must be overwritten")
    
    def isDifferent(self, other):
        """Function to check if two vectors are identical"""
        
        raise NotImplementedError("isDifferent must be overwritten")
    
    def clipVector(self, low, high):
        """
           Function to bound vector values based on input vectors min and max
        """
        raise NotImplementedError("clipVector must be overwritten")
    
    def plot(self):
        """Method to get a plottable array"""
        return self.getNdArray()


# Set of vectors (useful to store results and same-Space vectors together)
class VectorSet:
    """Class to store different vectors that live in the same Space"""
    
    def __init__(self):
        """Default constructor"""
        self.vecSet = []  # List of vectors of the set
    
    def __del__(self):
        """Default destructor"""
    
    def append(self, other, copy=True):
        """Method to add vector to the set"""
        # Checking dimensionality if a vector is present
        if self.vecSet:
            if not self.vecSet[0].checkSame(other):
                raise ValueError("ERROR! Provided vector not in the same Space of the vector set")
        
        self.vecSet.append(other.clone()) if copy else self.vecSet.append(other)
    
    def writeSet(self, filename, mode="a"):
        """Method to write to SEPlib file (by default it appends vectors to file)"""
        if mode not in "aw":
            raise ValueError("ERROR! mode must be either a (append) or w (write)")
        for idx_vec, vec_i in enumerate(self.vecSet):
            if mode == "w" and idx_vec == 0:
                wr_mode = "w"
            else:
                wr_mode = "a"
            vec_i.writeVec(filename, wr_mode)
        self.vecSet = []  # List of vectors of the set
    

class superVector(Vector):
    
    def __init__(self, *args):
        """
        superVector constructor
        :param args: vectors or superVectors or vectors list objects
        """
        self.vecs = []
        for v in args:
            if v is None:
                continue
            elif isinstance(v, Vector):
                self.vecs.append(v)
            elif isinstance(v, list):
                for vv in v:
                    if vv is None:
                        continue
                    elif isinstance(vv, Vector):
                        self.vecs.append(vv)
            else:
                raise TypeError('Argument must be either a vector or a superVector')
        
        self.n = len(self.vecs)
        # super(superVector, self).__init__()
        # self.shape = None
        # self.size = None
        # self.ndim = None
        
    @property
    def ndim(self):
        return [self.vecs[idx].ndim for idx in range(self.n)]
    
    @property
    def shape(self):
        return [self.vecs[idx].shape for idx in range(self.n)]
    
    @property
    def size(self):
        return sum([self.vecs[idx].size for idx in range(self.n)])
    
    def __del__(self):
        """superVector destructor"""
        del self.vecs, self.n
    
    def getNdArray(self):
        """Function to return Ndarray of the vector"""
        return [self.vecs[idx].getNdArray() for idx in range(self.n)]
    
    @property
    def shape(self):
        return [self.vecs[idx].shape for idx in range(self.n)]
    
    @property
    def size(self):
        return sum([self.vecs[idx].size for idx in range(self.n)])
    
    def norm(self, N=2):
        """Function to compute vector N-norm"""
        norm = [pow(self.vecs[idx].norm(N), N) for idx in range(self.n)]
        return pow(sum(norm), 1. / N)
    
    def set(self, val):
        """Function to set all values in the vector"""
        for idx in range(self.n):
            self.vecs[idx].set(val)
        return self
    
    def zero(self):
        """Function to zero out a vector"""
        for idx in range(self.n):
            self.vecs[idx].zero()
        return self
    
    def max(self):
        """Function to obtain maximum value within a vector"""
        return max([self.vecs[idx].max() for idx in range(self.n)])
    
    def min(self):
        """Function to obtain minimum value within a vector"""
        return min([self.vecs[idx].min() for idx in range(self.n)])
    
    def scale(self, sc):
        """Function to scale a vector"""
        if type(sc) is not list:
            sc = [sc] * self.n
        for idx in range(self.n):
            self.vecs[idx].scale(sc[idx])
        return self
    
    def addbias(self, bias):
        """Add a constant to the vector"""
        if type(bias) is not list:
            bias = [bias] * self.n
        for idx in range(self.n):
            self.vecs[idx].addbias(bias[idx])
        return self
    
    def rand(self, snr=1.0):
        """Function to randomize a vector"""
        for idx in range(self.n):
            self.vecs[idx].rand()
        return self
    
    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        vecs = [self.vecs[idx].clone() for idx in range(self.n)]
        return superVector(vecs)
    
    def cloneSpace(self):
        """Function to clone vector space"""
        return superVector([self.vecs[idx].cloneSpace() for idx in range(self.n)])
    
    def checkSame(self, other):
        """Function to check to make sure the vectors exist in the same space"""
        # Checking type
        if not isinstance(other, superVector):
            raise TypeError('Input variable is not a superVector')
        checkspace = np.asarray([self.vecs[idx].checkSame(other.vecs[idx]) for idx in range(self.n)])
        notsame = np.where(checkspace is False)[0]
        for v in notsame:
            raise Warning('Component %d not in the same space!' % v)
        return np.all(checkspace == True)
    
    # Combination of different vectors
    def copy(self, vecs_in):
        """Function to copy vector from input vector"""
        # Checking type
        if type(vecs_in) is not superVector:
            raise TypeError("Input variable is not a superVector")
        # Checking dimensionality
        if not self.checkSame(vecs_in):
            raise ValueError("ERROR! Dimensionality mismatching between given superVectors")
        for idx in range(self.n):
            self.vecs[idx].copy(vecs_in.vecs[idx])
        return self
    
    def scaleAdd(self, vecs_in, sc1=1.0, sc2=1.0):
        """Function to scale input vectors and add them to the original ones"""
        # Checking type
        if type(vecs_in) is not superVector:
            raise TypeError("Input variable is not a superVector")
        # Checking dimensionality
        if not self.checkSame(vecs_in):
            raise ValueError("ERROR! Dimensionality mismatching between given superVectors")
        for idx in range(self.n):
            self.vecs[idx].scaleAdd(vecs_in.vecs[idx], sc1, sc2)
        return self
    
    def dot(self, vecs_in):
        """Function to compute dot product between two vectors"""
        # Checking type
        if type(vecs_in) is not superVector:
            raise TypeError("Input variable is not a superVector")
        # Checking dimensionality
        if not self.checkSame(vecs_in):
            raise ValueError("ERROR! Dimensionality mismatching between given superVectors")
        return sum([self.vecs[idx].dot(vecs_in.vecs[idx]) for idx in range(self.n)])
    
    def multiply(self, vecs_in):
        """Function to multiply element-wise two vectors"""
        # Checking type
        if type(vecs_in) is not superVector:
            raise TypeError("Input variable is not a superVector")
        # Checking dimensionality
        if not self.checkSame(vecs_in):
            raise ValueError("ERROR! Dimensionality mismatching between given superVectors")
        for idx in range(self.n):
            self.vecs[idx].multiply(vecs_in.vecs[idx])
        return self
    
    def isDifferent(self, vecs_in):
        """Function to check if two vectors are identical"""
        # Checking type
        if type(vecs_in) is not superVector:
            raise TypeError("Input variable is not a superVector")
        return any([self.vecs[idx].isDifferent(vecs_in.vecs[idx]) for idx in range(self.n)])
    
    def clipVector(self, lows, highs):
        for idx in range(self.n):
            self.vecs[idx].clipVector(lows[idx], highs[idx])
        return self
    
    def abs(self):
        for idx in range(self.n):
            self.vecs[idx].abs()
        return self
    
    def sign(self):
        for idx in range(self.n):
            self.vecs[idx].sign()
        return self
    
    def reciprocal(self):
        for idx in range(self.n):
            self.vecs[idx].reciprocal()
        return self
    
    def maximum(self, other):
        if np.isscalar(other):
            for idx in range(self.n):
                self.vecs[idx].maximum(other)
            return self
        elif type(other) is not superVector:
            raise TypeError("Input variable is not a superVector")
        if other.n != self.n:
            raise ValueError('Input must have the same length of self')
        for idx in range(self.n):
            self.vecs[idx].maximum(other.vecs[idx])
        return self
    
    def conj(self):
        for idx in range(self.n):
            self.vecs[idx].conj()
        return self
    
    def real(self):
        for idx in range(self.n):
            self.vecs[idx].real()
        return self
    
    def imag(self, ):
        for idx in range(self.n):
            self.vecs[idx].imag()
        return self
    
    def pow(self, power):
        for idx in range(self.n):
            self.vecs[idx].pow(power)
        return self
    
    def writeVec(self, filename, mode="w"):
        """Method to write to vector to file within a Vector set"""
        _, ext = os.path.splitext(filename)
        for ii, vec_cmp in enumerate(self.vecs):
            # Writing components to different files
            filename_cmp = ".".join(filename.split('.')[:-1]) + "_comp%s" % (ii + 1)
            # Writing files (recursively)
            vec_cmp.writeVec(filename_cmp + ext, mode)
        return



from collections import deque
from math import isnan
import numpy as np
from copy import deepcopy

from occamypy import operator as O
from occamypy import problem as P
from occamypy import solver as S

# Check for avoid Overflow or Underflow
zero = 10 ** (np.floor(np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)


# Beta functions
# grad=new gradient, grad0=old, dir=search direction
# From A SURVEY OF NONLINEAR CONJUGATE GRADIENT METHODS

def _betaFR(grad, grad0, dir, logger):
    """Fletcher and Reeves method"""
    # betaFR = sum(dprod(g,g))/sum(dprod(g0,g0))
    dot_grad = grad.dot(grad)
    dot_grad0 = grad0.dot(grad0)
    if dot_grad0 == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of previous gradient is zero!!!")
    else:
        beta = dot_grad / dot_grad0
    return beta


def _betaPRP(grad, grad0, dir, logger):
    """Polak, Ribiere, Polyak method"""
    # betaPRP = sum(dprod(g,g-g0))/sum(dprod(g0,g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    dot_num = tmp1.dot(grad)
    dot_grad0 = grad0.dot(grad0)
    if dot_grad0 == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of previous gradient is zero!!!")
    else:
        beta = dot_num / dot_grad0
    return beta


def _betaHS(grad, grad0, dir, logger):
    """Hestenes and Stiefel"""
    # betaHS = sum(dprod(g,g-g0))/sum(dprod(d,g-g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    dot_num = tmp1.dot(grad)
    dot_denom = tmp1.dot(dir)
    if dot_denom == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        beta = dot_num / dot_denom
    return beta


def _betaCD(grad, grad0, dir, logger):
    """Conjugate Descent"""
    # betaCD = -sum(dprod(g,g))/sum(dprod(d,g0))
    dot_num = grad.dot(grad)
    dot_denom = -grad0.dot(dir)
    if dot_denom == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        beta = dot_num / dot_denom
    return beta


def _betaLS(grad, grad0, dir, logger):
    """Liu and Storey"""
    # betaLS = -sum(dprod(g,g-g0))/sum(dprod(d,g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    dot_num = tmp1.dot(grad)
    dot_denom = -grad0.dot(dir)
    if dot_denom == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        beta = dot_num / dot_denom
    return beta


def _betaDY(grad, grad0, dir, logger):
    """Dai and Yuan"""
    # betaDY = sum(dprod(g,g))/sum(dprod(d,g-g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    dot_num = grad.dot(grad)
    dot_denom = tmp1.dot(dir)
    if dot_denom == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        beta = dot_num / dot_denom
    return beta


def _betaBAN(grad, grad0, dir, logger):
    """Bamigbola, Ali and Nwaeze"""
    # betaDY = sum(dprod(g,g-g0))/sum(dprod(g0,g-g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    dot_num = tmp1.dot(grad)
    dot_denom = tmp1.dot(grad0)
    if dot_denom == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        beta = -dot_num / dot_denom
    return beta


def _betaHZ(grad, grad0, dir, logger):
    """Hager and Zhang"""
    # betaN = sum(dprod(g-g0-2*sum(dprod(g-g0,g-g0))*d/sum(dprod(d,g-g0)),g))/sum(dprod(d,g-g0))
    tmp1 = grad.clone()
    # g-g0
    tmp1.scaleAdd(grad0, 1.0, -1.0)
    # sum(dprod(g-g0,g-g0))
    dot_diff_g_g0 = tmp1.dot(tmp1)
    # sum(dprod(d,g-g0))
    dot_dir_diff_g_g0 = tmp1.dot(dir)
    if dot_dir_diff_g_g0 == 0.:  # Avoid division by zero
        beta = 0.
        if logger:
            logger.addToLog("Setting beta to zero since norm of denominator is zero!!!")
    else:
        # g-g0-2*sum(dprod(g-g0,g-g0))*d/sum(dprod(d,g-g0))
        tmp1.scaleAdd(dir, 1.0, -2.0 * dot_diff_g_g0 / dot_dir_diff_g_g0)
        # sum(dprod(g-g0-2*sum(dprod(g-g0,g-g0))*d/sum(dprod(d,g-g0)),g))
        dot_num = grad.dot(tmp1)
        # dot_num/sum(dprod(d,g-g0))
        beta = dot_num / dot_dir_diff_g_g0
    return beta


def _betaSD(grad, grad0, dir, logger):
    """Steepest descent"""
    beta = 0.
    return beta


class NLCG(S.Solver):
    """Non-Linear Conjugate Gradient and Steepest-Descent Solver object"""
    
    # Default class methods/functions
    def __init__(self, stoppr, stepper=None, beta_type="FR", logger=None):
        """Constructor for NLCG Solver"""
        # Calling parent construction
        super(NLCG, self).__init__()
        # Defining stopper object
        self.stoppr = stoppr
        
        # Defining stepper object
        self.stepper = stepper if stepper is not None else S.ParabolicStep()
        # Beta function to use during the inversion
        self.beta_type = beta_type
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stoppr.logger = self.logger
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d, geval = %d"
        return
    
    def __del__(self):
        """Default destructor"""
        return
    
    def beta_func(self, grad, grad0, dir):
        """Beta function interface"""
        beta_type = self.beta_type
        if beta_type == "FR":
            beta = _betaFR(grad, grad0, dir, self.logger)
        elif beta_type == "PRP":
            beta = _betaPRP(grad, grad0, dir, self.logger)
        elif beta_type == "HS":
            beta = _betaHS(grad, grad0, dir, self.logger)
        elif beta_type == "CD":
            beta = _betaCD(grad, grad0, dir, self.logger)
        elif beta_type == "LS":
            beta = _betaLS(grad, grad0, dir, self.logger)
        elif beta_type == "DY":
            beta = _betaDY(grad, grad0, dir, self.logger)
        elif beta_type == "BAN":
            beta = _betaBAN(grad, grad0, dir, self.logger)
        elif beta_type == "HZ":
            beta = _betaHZ(grad, grad0, dir, self.logger)
        elif beta_type == "SD":
            beta = _betaSD(grad, grad0, dir, self.logger)
        else:
            raise ValueError("ERROR! Requested Beta function type not existing")
        return beta
    
    def run(self, problem, verbose=False, restart=False):
        """Running NLCG or steppest-descent solver"""
        
        self.create_msg = verbose or self.logger
        
        # Resetting stopper before running the inversion
        self.stoppr.reset()
        
        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "\t\t\tNON-LINEAR %s SOLVER log file\n" % (
                    "STEEPEST-DESCENT" if self.beta_type == "SD" else "CONJUGATE GRADIENT")
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                if self.beta_type != "SD":
                    msg += "\tConjugate method used: %s\n" % self.beta_type
                msg += 90 * "#" + "\n"
                if verbose:
                    print(msg.replace("log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
            
            # Setting internal vectors (model, search direction, and previous gradient vectors)
            prblm_mdl = problem.get_model()
            cg_mdl = prblm_mdl.clone()
            cg_dmodl = prblm_mdl.clone()
            cg_dmodl.zero()
            cg_grad0 = cg_dmodl.clone()
            
            # Other internal variables
            beta = 0.0
            iiter = 0
        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            self.stepper.alpha = self.restart.retrieve_parameter("alpha")
            initial_obj_value = self.restart.retrieve_parameter(
                "obj_initial")  # Retrieving initial objective function value
            cg_mdl = self.restart.retrieve_vector("cg_mdl")
            cg_dmodl = self.restart.retrieve_vector("cg_dmodl")
            cg_grad0 = self.restart.retrieve_vector("cg_grad0")
            # Setting the model and residuals to avoid residual twice computation
            problem.set_model(cg_mdl)
            prblm_mdl = problem.get_model()
            # Setting residual vector to avoid its unnecessary computation
            problem.set_residual(self.restart.retrieve_vector("prblm_res"))
        
        # Common variables unrelated to restart
        prev_mdl = prblm_mdl.clone().zero()
        
        while True:
            # Computing objective function
            obj0 = problem.get_obj(cg_mdl)  # Compute objective function value
            prblm_res = problem.get_res(cg_mdl)  # Compute residuals
            prblm_grad = problem.get_grad(cg_mdl)  # Compute the gradient
            if iiter == 0:
                initial_obj_value = obj0  # For relative objective function value
                # Saving objective function value
                self.restart.save_parameter("obj_initial", initial_obj_value)
                # iteration info
                if self.create_msg:
                    msg = self.iter_msg % (str(iiter).zfill(self.stoppr.zfill),
                                           obj0,
                                           problem.get_rnorm(cg_mdl),
                                           problem.get_gnorm(cg_mdl),
                                           problem.get_fevals(),
                                           problem.get_gevals())
                    # Writing on log file
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0) or isnan(prblm_grad.norm()):
                    raise ValueError("ERROR! Either gradient norm or objective function value NaN!")
            if prblm_grad.norm() == 0.:
                print("Gradient vanishes identically")
                break
            
            # Saving results
            self.save_results(iiter, problem, force_save=False)
            # Keeping current inverted model
            prev_mdl.copy(prblm_mdl)
            
            if iiter >= 1:
                beta = self.beta_func(prblm_grad, cg_grad0, cg_dmodl)
                if beta < 0.:
                    if self.logger:
                        self.logger.addToLog("Beta negative setting to zero: beta value=%s!!!" % beta)
                    beta = 0.
            if self.beta_type != "SD":
                if self.logger:
                    self.logger.addToLog("beta coefficient: %s" % beta)
            
            # dmodl = beta*dmodl - grad
            cg_dmodl.scaleAdd(prblm_grad, beta, -1.0)
            # grad0 = grad
            cg_grad0.copy(prblm_grad)
            # Calling line search
            alpha, success = self.stepper.run(problem, cg_mdl, cg_dmodl, self.logger)
            if not success:
                if self.create_msg:
                    msg = "Stepper couldn't find a proper step size, will terminate solver"
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break
            
            # Increasing iteration counter
            iiter = iiter + 1
            obj1 = problem.get_obj(cg_mdl)  # Compute objective function value
            # Redundant tests on verifying convergence
            if obj0 <= obj1:
                if self.create_msg:
                    msg = "Objective function at new point greater or equal than previous one:\n\t" \
                          "obj_new = %.5e\tobj_cur = %.5e\n" \
                          "Potential issue in the stepper or in revaluation of objective function! Solver will stop!" \
                          % (obj1, obj0)
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break
            
            # Saving current model and previous search direction in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_parameter("alpha", alpha)
            self.restart.save_vector("cg_mdl", cg_mdl)
            self.restart.save_vector("cg_dmodl", cg_dmodl)
            self.restart.save_vector("cg_grad0", cg_grad0)
            # Saving data space vectors
            self.restart.save_vector("prblm_res", prblm_res)
            # iteration info
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stoppr.zfill),
                                       obj1,
                                       problem.get_rnorm(cg_mdl),
                                       problem.get_gnorm(cg_mdl),
                                       problem.get_fevals(),
                                       problem.get_gevals())
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1) or isnan(prblm_grad.norm()):
                raise ValueError("ERROR! Either gradient norm or objective function value NaN!")
            if self.stoppr.run(problem, iiter, initial_obj_value, verbose):
                break
        
        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "\t\t\tNON-LINEAR %s SOLVER log file end\n" % (
                "STEEPEST-DESCENT" if self.beta_type == "SD" else "CONJUGATE GRADIENT")
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
        
        # Clear restart object
        self.restart.clear_restart()
        
        return


class TNewton(S.Solver):
    """Truncated Newton/Gauss-Newton solver object"""
    
    def __init__(self, stopper, niter_max, HessianOp, stepper=None, niter_min=None, warm_start=True, Newton_prefix=None,
                 logger=None):
        """
        Constructor for Truncated-Newton Solver.
        stoppr     = [no default] - stopper class; Stopper object necessary to terminate the solver
        niter_max  = [no default] - int; Maximum number of iterations for solving Newton system when starting with a zero initial model
        HessianOp  = [no default] - operator class; Operator to apply Hessian matrix onto model vector. The operator must contain a set_background function to set model vector on which the Hessian is computed. Note the symmetric solver will be used. Hence, this operator must be symmetric.
        stepper    = [CvSrch] - stepper class; Stepper object necessary to perform line-search step
        niter_min  = [niter_max] - int; Number of iterations for solving Newton system when linear inversion starts from previous search direction
        warm_start = [False] - boolean; If True, the linear Hessian inversion is started from the previous search direction if aligned with the current gradient
        logger 	  = [None] - logger class; Logger object to save inversion information at runtime
        """
        # Defining stopper
        self.stopper = stopper  # Stopper for non-linear problem
        # Setting maximum and minimum number of iterations
        self.niter_max = niter_max
        self.niter_min = niter_max
        # Setting linear inversion iterations
        if niter_min is not None:
            if niter_min <= niter_max:
                raise ValueError(
                    "niter_min of %d must be smaller or equal than niter_max of %d." % (niter_min, niter_max))
            self.niter_min = niter_min
        # Defining stepper object
        self.stepper = stepper if stepper is not None else S.CvSrchStep()
        # Warm starts requested?
        self.warm_start = warm_start
        # Hessian operator
        if HessianOp is not None:
            if "set_background" not in dir(HessianOp):
                raise AttributeError("Hessian operator must have a set_background function.")
        # Setting linear solver for solving Newton system and problem class
        StopLin = S.BasicStopper(niter=self.niter_max)
        self.lin_solver = S.CGsym(StopLin)
        self.NewtonPrblm = P.LeastSquaresSymmetric(HessianOp.domain.clone(), HessianOp.domain.clone(), HessianOp)
        return
    
    def run(self, problem, verbose=False, restart=False):
        """Running Truncated Newton solver"""
        return


class LBFGS(S.Solver):
    """L-BFGS (Limited-memory Broyden-Fletcher-Goldfarb-Shanno) Solver object"""
    
    def __init__(self, stopper, stepper=None, save_alpha=False, m_steps=None, H0=None, logger=None, save_est=False):
        """
		Constructor for LBFGS Solver:
		:param stopper    : Stopper, object to terminate the solver
		:param stepper    : Stepper, object to perform line-search step
		:param save_alpha : bool, Use previous step-length value as initial guess. Otherwise, the algorithm starts from an initial guess of 1.0 [False]
		:param m_steps    : int, Maximum number of steps to store to estimate the inverse Hessian (by default it runs BFGS method)
		:param H0         : Operator, initial estimated Hessian inverse (by default it assumes an identity operator)
		:param logger 	  : Logger, object to save inversion information at runtime
		:param save_est   : bool, save inverse Hessian estimate vectors (self.prefix must not be None) [False]
		"""
        # Calling parent construction
        super(LBFGS, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Defining stepper object
        self.stepper = stepper if stepper is not None else S.CvSrchStep()
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # LBFGS-specific parameters
        self.save_alpha = save_alpha
        self.H0 = H0
        self.m_steps = m_steps
        self.save_est = save_est
        self.tmp_vector = None  # A copy of the model vector will be create when the function run is invoked
        self.iistep = 0  # necessary to re-used the estimated hessian inverse from previous runs
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d, geval = %d"
    
    def save_hessian_estimate(self, index, iiter):
        """Function to save current vector of estimated Hessian inverse"""
        # index of the step to save
        if self.prefix is not None and self.save_est:
            step_filename = self.prefix + "step_vector_%s.H" % iiter
            grad_diff_filename = self.prefix + "grad_diff_vector_%s.H" % iiter
            self.step_vectors[index].writeVec(step_filename)
            self.grad_diff_vectors[index].writeVec(grad_diff_filename)
    
    def check_rho(self, denom_dot, step_index, iiter):
        """Function to check scaling factor of Hessian inverse estimate"""
        if denom_dot == 0.:
            if self.m_steps is not None:
                self.rho[step_index] = 0.
            else:
                self.rho.append(0.)
            msg = "Skipping update to estimated Hessian; y vector orthogonal to s vector at iteration %s" % iiter
            if self.logger:
                self.logger.addToLog(msg)
        elif denom_dot < 0.:
            if self.m_steps is not None:
                self.rho[step_index] = 0.
            else:
                self.rho.append(0.)
            msg = "Skipping update to estimated Hessian; not positive at iteration %s" % iiter
            if self.logger:
                self.logger.addToLog(msg)
        else:
            if self.m_steps is not None:
                self.rho[step_index] = 1.0 / denom_dot
            else:
                self.rho.append(1.0 / denom_dot)
            # Saving current update for inverse Hessian estimate (i.e., gradient-difference and model-step vectors)
            self.save_hessian_estimate(step_index, iiter)
        return
    
    # BFGSMultiply function
    def BFGSMultiply(self, dmodl, grad, iiter):
        """Function to apply approximated inverse Hessian"""
        # Array containing dot-products
        if self.m_steps is not None:
            alpha = [0.0] * self.m_steps
            # Handling of limited memory
            if iiter <= self.m_steps:
                initial_point = 0
            else:
                initial_point = iiter % self.m_steps
            # Right step list
            rloop = deque(range(0, min(iiter, self.m_steps)))
            rloop.reverse()
            # Rotate the list
            rloop.rotate(initial_point)
            # Left step list
            lloop = deque(range(0, min(iiter, self.m_steps)))
            # Rotate the list
            lloop.rotate(-initial_point)
        else:
            alpha = [0.0] * iiter
            rloop = deque(range(0, iiter))
            rloop.reverse()
            lloop = deque(range(0, iiter))
        # r = -grad
        dmodl.copy(grad)
        dmodl.scale(-1.0)
        # Apply right-hand series of operators
        for ii in rloop:
            # Check positivity, if not true skip the update
            if self.rho[ii] > 0.0:
                # alpha_i=rho_i*s_i'r
                alpha[ii] = self.rho[ii] * self.step_vectors[ii].dot(dmodl)
                # r=r-alpha_i*y_i
                dmodl.scaleAdd(self.grad_diff_vectors[ii], 1.0, -alpha[ii])
        # Comput center (If not provide Identity matrix is assumed)
        # r=H0r
        if self.H0 is not None:
            # Apply a forward of the initial Hessian estimate
            self.H0.forward(False, dmodl, self.tmp_vector)
            dmodl.copy(self.tmp_vector)
        # Apply left-hand series of operators
        for ii in lloop:
            # Check positivity, if not true skip the update
            if self.rho[ii] > 0.0:
                # beta=rhoiyi'r
                beta = self.rho[ii] * self.grad_diff_vectors[ii].dot(dmodl)
                dmodl.scaleAdd(self.step_vectors[ii], 1.0, alpha[ii] - beta)
        return
    
    def run(self, problem, verbose=False, keep_hessian=False, restart=False):
        """
        Running LBFGS solver
        :param problem: problem to be minimized
        :param verbose: verbosity flag [False]
        :param keep_hessian: use hessian inverse estimate build from previous runs [False]
        :param restart: restart previously crashed inversion [False]
        """
        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Getting model vector
        prblm_mdl = problem.get_model()
        # Preliminary variables for Hessian inverse estimation
        if not keep_hessian or "rho" not in dir(self):
            if self.m_steps is not None:
                self.step_vectors = [None] * self.m_steps  # s_i vectors
                self.grad_diff_vectors = [None] * self.m_steps  # y_i vectors
                self.rho = [None] * self.m_steps  # Scalar term necessary for Hessian inverse estimation
            else:
                self.step_vectors = []  # s_i vectors
                self.grad_diff_vectors = []  # y_i vectors
                self.rho = []  # Scalar term necessary for Hessian inverse estimation
            self.iistep = 0
        
        if not restart:
            msg = 90 * "#" + "\n"
            if self.m_steps is not None:
                msg += "Limited-memory Broyden-Fletcher-Goldfarb-Shanno (L-BFGS) algorithm log file\n"
                msg += "Maximum number of steps to be used for Hessian inverse estimation: %s \n" % self.m_steps
            else:
                msg += "Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm log file\n"
            # Printing restart folder
            msg += "Restart folder: %s\n" % self.restart.restart_folder
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace("log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
            
            # Setting internal vectors (model, search direction, and previous gradient vectors)
            bfgs_mdl = prblm_mdl.clone()
            bfgs_dmodl = prblm_mdl.clone()
            bfgs_dmodl.zero()
            bfgs_grad0 = bfgs_dmodl.clone()
            
            # Other internal variables
            iiter = 0
        else:
            # Retrieving parameters and vectors to restart the solver
            msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            self.stepper.alpha = self.restart.retrieve_parameter("alpha")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            bfgs_mdl = self.restart.retrieve_vector("bfgs_mdl")
            bfgs_dmodl = self.restart.retrieve_vector("bfgs_dmodl")
            bfgs_grad0 = self.restart.retrieve_vector("bfgs_grad0")
            # Setting the model and residuals to avoid residual twice computation
            problem.set_model(bfgs_mdl)
            prblm_mdl = problem.get_model()
            # Setting residual vector to avoid its unnecessary computation
            problem.set_residual(self.restart.retrieve_vector("prblm_res"))
            # Retrieving Hessian inverse estimate
            self.rho = self.restart.retrieve_parameter("rho")
            self.iistep = self.restart.retrieve_parameter("iistep")
            for istep in range(min(iiter, self.iistep)):
                if self.m_steps is not None:
                    if istep < self.m_steps:
                        self.grad_diff_files[istep] = self.restart.retrieve_vector("grad_diff_vectors%s" % istep)
                        self.step_files[istep] = self.restart.retrieve_vector("step_vectors%s.H" % istep)
                else:
                    self.grad_diff_files.append(self.restart.retrieve_vector("grad_diff_vectors%s" % istep))
                    self.step_files.append(self.restart.retrieve_vector("step_vectors%s" % istep))
        
        # Common variables unrelated to restart
        self.tmp_vector = bfgs_dmodl.clone()
        self.tmp_vector.zero()
        prev_mdl = prblm_mdl.clone().zero()
        
        # Inversion loop
        while True:
            # Computing objective function
            obj0 = problem.get_obj(bfgs_mdl)  # Compute objective function value
            prblm_res = problem.get_res(bfgs_mdl)  # Compute residuals
            prblm_grad = problem.get_grad(bfgs_mdl)  # Compute the gradient
            if iiter == 0:
                initial_obj_value = obj0  # For relative objective function value
                # Saving objective function value
                self.restart.save_parameter("obj_initial", initial_obj_value)
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj0,
                                       problem.get_rnorm(bfgs_mdl),
                                       problem.get_gnorm(bfgs_mdl),
                                       problem.get_fevals(),
                                       problem.get_gevals())
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0) or isnan(prblm_grad.norm()):
                    raise ValueError("Either gradient norm or objective function value NaN!")
            if prblm_grad.norm() == 0.:
                print("Gradient vanishes identically")
                break
            
            # Saving results
            self.save_results(iiter, problem, force_save=False)
            # Saving current inverted model
            prev_mdl.copy(prblm_mdl)
            
            # Applying approximated Hessian inverse
            msg = "Appplying inverse Hessian estimate"
            if self.m_steps is not None:
                msg += "\nCurrent inverse dot-products of BFGS estimation vectors %s" \
                       % (self.rho[0:min(self.m_steps, self.iistep)])
            else:
                if len(self.rho) > 0:
                    msg += "\nCurrent inverse dot-products of BFGS estimation vectors %s" % self.rho
            if self.logger:
                self.logger.addToLog(msg)
            self.BFGSMultiply(bfgs_dmodl, prblm_grad, self.iistep)
            msg = "Done applying inverse Hessian estimate"
            if self.logger:
                self.logger.addToLog(msg)
            
            # grad0 = grad
            bfgs_grad0.copy(prblm_grad)
            # Calling line search
            alpha, success = self.stepper.run(problem, bfgs_mdl, bfgs_dmodl, self.logger)
            if not success:
                msg = "Stepper couldn't find a proper step size, will terminate solver"
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break
            
            obj1 = problem.get_obj(bfgs_mdl)  # Compute objective function value
            # Redundant tests on verifying convergence
            if obj0 <= obj1:
                msg = "Objective function at new point greater or equal than previous one: obj_fun_old=%s obj_fun_new=%s\n" \
                      "Potential issue in the stepper or in revaluation of objective function!" % (obj0, obj1)
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                raise ValueError(msg)
            
            # Compute new gradient
            prblm_grad = problem.get_grad(bfgs_mdl)
            # Compute updates for estimated Hessian inverse
            if self.m_steps is not None:
                # LBFGS
                step_index = self.iistep % self.m_steps  # Modulo to handle limited memory
                # yn+1=gn+1-gn
                self.grad_diff_vectors[step_index] = bfgs_grad0.clone()
                self.grad_diff_vectors[step_index].scaleAdd(prblm_grad, -1.0, 1.0)
                # sn+1=xn+1-xn = alpha * dmodl
                self.step_vectors[step_index] = bfgs_dmodl.clone()
                self.step_vectors[step_index].scale(alpha)
            else:
                # BFGS
                step_index = self.iistep
                # yn+1=gn+1-gn
                self.grad_diff_vectors.append(bfgs_grad0.clone())
                self.grad_diff_vectors[step_index].scaleAdd(prblm_grad, -1.0, 1.0)
                # sn+1=xn+1-xn = alpha * dmodl
                self.step_vectors.append(bfgs_dmodl.clone())
                self.step_vectors[step_index].scale(alpha)
            # rhon+1=1/yn+1'sn+1
            denom_dot = self.grad_diff_vectors[step_index].dot(self.step_vectors[step_index])
            # Checking rho
            self.check_rho(denom_dot, step_index, self.iistep)
            
            # Making first step-length value Hessian guess if not provided by user
            if iiter == 0 and alpha != 1.0:
                self.restart.save_parameter("fist_alpha", alpha)
                self.H0 = O.Scaling(bfgs_dmodl, alpha) if self.H0 is None else self.H0 * O.Scaling(bfgs_dmodl, alpha)
                if self.logger:
                    self.logger.addToLog("First step-length value added to first Hessian inverse estimate!")
                self.stepper.alpha = 1.0
            
            # Increasing iteration counter
            iiter = iiter + 1
            self.iistep += 1
            
            # Using alpha = 1.0 after first iteration
            if iiter != 0 and not self.save_alpha:
                self.stepper.alpha = 1.0
            
            # Saving current model and previous search direction in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_parameter("alpha", alpha)
            self.restart.save_vector("bfgs_mdl", bfgs_mdl)
            self.restart.save_vector("bfgs_dmodl", bfgs_dmodl)
            self.restart.save_vector("bfgs_grad0", bfgs_grad0)
            # Saving Inverse Hessian estimate for restart
            self.restart.save_parameter("rho", self.rho)
            self.restart.save_parameter("iistep", self.iistep)
            self.restart.save_vector("grad_diff_vectors%s" % step_index, self.grad_diff_vectors[step_index])
            self.restart.save_vector("step_vectors%s" % step_index, self.step_vectors[step_index])
            # Saving data space vectors
            self.restart.save_vector("prblm_res", prblm_res)
            
            # iteration info
            msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                   obj1,
                                   problem.get_rnorm(bfgs_mdl),
                                   problem.get_gnorm(bfgs_mdl),
                                   problem.get_fevals(),
                                   problem.get_gevals())
            if verbose:
                print(msg)
            # Writing on log file
            if self.logger:
                self.logger.addToLog("\n" + msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1) or isnan(prblm_grad.norm()):
                raise ValueError("Either gradient norm or objective function value NaN!")
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break
        
        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        msg = 90 * "#" + "\n"
        if self.m_steps is not None:
            msg += "Limited-memory Broyden-Fletcher-Goldfarb-Shanno (L-BFGS) algorithm log file end\n"
        else:
            msg += "Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm log file end\n"
        msg += 90 * "#" + "\n"
        if verbose:
            print(msg.replace("log file ", ""))
        if self.logger:
            self.logger.addToLog(msg)
        self.restart.clear_restart()
        # Resetting inverse Hessian matrix
        if not keep_hessian:
            self.H0 = None
        del self.tmp_vector
        self.tmp_vector = None


class LBFGSB(S.Solver):
    """L-BFGS-B (Limited-memory Broyden-Fletcher-Goldfarb-Shanno with Bounds) Solver object
       Implementation inspired by the one in the GitHub repo: https://github.com/bgranzow/L-BFGS-B.git
    """

    def __init__(self, stopper, stepper=None, m_steps=np.inf, logger=None):
        """
		Constructor for LBFGS-B Solver:
		:param stopper    : Stopper, object to terminate the solver
		:param stepper    : Stepper, object to perform line-search step
		:param m_steps    : int, Maximum number of steps to store to estimate the inverse Hessian (by default it runs BFGS method)
		:param logger 	  : Logger, object to save inversion information at runtime
		"""
        # Calling parent construction
        super(LBFGSB, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Defining stepper object
        self.stepper = stepper if stepper is not None else S.StrongWolfe()
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # LBFGSB-specific parameters
        self.m_steps = m_steps
        self.epsmch = None
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d, geval = %d"

    def get_breakpoints(self, bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound):
        """
        Compute the breakpoint variables needed for the Cauchy point.
        Equations (4.1),(4.2), and F in Algorigthm CP: Initialize.
        The input vector bfgsb_dmodl will contain the search direction
        :return:
        :t_vec breakpoint vector
        :F the indices that sort t from low to high
        """
        n_mod = bfgsb_mdl.size
        t_vec = bfgsb_mdl.clone().zero()
        bfgsb_dmodl.copy(prblm_grad)
        bfgsb_dmodl.scale(-1.0)
        # Getting NdArrays of variables
        g = prblm_grad.getNdArray().ravel()
        x = bfgsb_mdl.getNdArray().ravel()
        t = t_vec.getNdArray().ravel()
        l = minBound.getNdArray().ravel()
        u = maxBound.getNdArray().ravel()
        d = bfgsb_dmodl.getNdArray().ravel()
        # Largest representable number
        realmax = np.finfo(self.epsmch).max
        for ii in range(n_mod):
            if g[ii] < 0.0:
                t[ii] = (x[ii] - u[ii]) / g[ii]
            elif g[ii] > 0.0:
                t[ii] = (x[ii] - l[ii]) / g[ii]
            else:
                t[ii] = realmax
            if t[ii] < self.epsmch:
                d[ii] = 0.0
        F = np.argsort(t)
        return t_vec, F

    def get_cauchy_point(self, bfgsb_mdl_cauchy, bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound, W, M, theta):
        """
        Compute the generalized Cauchy point.
        Algorithm CP, Pages 8-9.
        :return:
        :c initialization vector for subspace minimization
        """
        t_vec, F = self.get_breakpoints(bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound)
        # xc = x;
        bfgsb_mdl_cauchy.copy(bfgsb_mdl)
        # Getting vector arrays
        d = bfgsb_dmodl.getNdArray().ravel()
        tt = t_vec.getNdArray().ravel()
        xc = bfgsb_mdl_cauchy.getNdArray().ravel()
        x = bfgsb_mdl.getNdArray().ravel()
        l = minBound.getNdArray().ravel()
        u = maxBound.getNdArray().ravel()
        g = prblm_grad.getNdArray().ravel()

        # Main algorithm
        p = np.matmul(W.T, d)
        c = np.zeros((W.shape[1]))
        fp = -bfgsb_dmodl.dot(bfgsb_dmodl)
        fpp = -theta * fp - np.sum(p * np.matmul(M, p))
        fpp0 = -theta * fp
        dt_min = -fp / (fpp + self.epsmch)
        t_old = 0
        for jj in range(bfgsb_mdl_cauchy.size):
            ii = jj
            if tt[F[ii]] > 0:
                break
        b = F[ii]
        t = tt[b]
        dt = t - t_old

        while dt_min > dt and ii < bfgsb_mdl_cauchy.size:
            if d[b] > 0.0:
                xc[b] = u[b]
            elif d[b] < 0.0:
                xc[b] = l[b]
            zb = xc[b] - x[b]
            c += dt * p
            gb = g[b]
            wbt = W[b,:]
            fp = fp + dt * fpp + gb * gb + theta * gb * zb - gb * np.sum(wbt * np.matmul(M, c))
            fpp = fpp - theta * gb * gb - 2.0 * gb * np.sum(wbt * np.matmul(M, p)) - gb * gb * np.sum(wbt * np.matmul(M, wbt))
            fpp = max(self.epsmch * fpp0, fpp)
            p += gb * wbt
            d[b] = 0.0
            dt_min = -fp / fpp
            t_old = t
            ii += 1
            if ii < bfgsb_mdl_cauchy.size:
                b = F[ii]
                t = tt[b]
                dt = t - t_old

        # Perform final updates
        dt_min = max(dt_min, 0)
        t_old = t_old + dt_min
        for jj in range(ii,bfgsb_mdl_cauchy.size):
            idx = F[jj]
            xc[idx] = x[idx] + t_old * d[idx]
        c += dt_min * p
        return c

    def find_alpha(self, l, u, xc, du, free_vars_idx):
        """
        Equation (5.8), Page 8.
        :return:
        :alpha_star positive scaling parameter.
        """
        alpha_star = 1.0
        n = len(free_vars_idx)
        for ii in range(n):
            idx = free_vars_idx[ii]
            if (du[ii] > 0.0):
                alpha_star = min(alpha_star, (u[idx] - xc[idx]) / du[ii])
            else:
                alpha_star = min(alpha_star, (l[idx] - xc[idx]) / (du[ii]+self.epsmch))
        ######## Check this!
        if alpha_star < 0.0:
            alpha_star = 0.0
        return alpha_star

    def subspace_min(self, bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound, bfgsb_mdl_cauchy, c, W, M, theta):
        """
        Subspace minimization for the quadratic model over free variables.
        Direct Primal Method, Page 12.
        :return:
        :free_var Flag whether there are free variables or not
        """
        # Setting the line search flag to true by default
        free_var = True

        # Getting arrays and parameters
        n_mod = bfgsb_mdl.size
        xc = bfgsb_mdl_cauchy.getNdArray().ravel()
        x = bfgsb_mdl.getNdArray().ravel()
        l = minBound.getNdArray().ravel()
        u = maxBound.getNdArray().ravel()
        g = prblm_grad.getNdArray().ravel()
        xbar = bfgsb_dmodl.getNdArray().ravel()

        # Finding free variables
        free_vars_idx = []
        Z = []
        for ii in range(n_mod):
            if xc[ii] != u[ii] and xc[ii] != l[ii]:
                free_vars_idx.append(ii)
                unit_tmp = np.zeros((n_mod))
                unit_tmp[ii] = 1.0
                Z.append(unit_tmp)
        Z = np.array(Z).T

        num_free_vars = len(free_vars_idx)

        if num_free_vars == 0:
            bfgsb_mdl.copy(bfgsb_mdl_cauchy) # xbar = x = xc
            free_var = False
            return free_var

        # Compute W^T Z, the restriction of W to free variables
        WTZ = np.matmul(W.T, Z)

        # Compute the reduced gradient of mk restricted to free variables
        rr = g + theta * (xc - x) - np.matmul(W, np.matmul(M, c))
        r = np.zeros((num_free_vars))
        for ii in range(num_free_vars):
            r[ii] = rr[free_vars_idx[ii]]

        # Forming intermediate variables
        invtheta = 1.0 / theta
        v = np.matmul(M, np.matmul(WTZ, r))
        N = invtheta * np.matmul(WTZ, WTZ.T)
        N = np.eye(N.shape[0]) - np.matmul(M, N)
        v = np.linalg.solve(N, v)
        du = - invtheta * ( r + invtheta * np.matmul(WTZ.T, v))

        # find alpha star
        alpha_star = self.find_alpha(l, u, xc, du, free_vars_idx)

        # compute the subspace minimization
        bfgsb_dmodl.copy(bfgsb_mdl_cauchy) # xbar = xc
        for ii in range(num_free_vars):
            idx = free_vars_idx[ii]
            xbar[idx] += alpha_star * du[ii]
        # Returning new search direction
        bfgsb_dmodl.scaleAdd(bfgsb_mdl,1.0,-1.0)

        return free_var

    def form_W(self, Y_mat, theta, S_mat):
        """
        :param Y_mat: list containing y_k vectors
        :param theta: theta parameters within the L-BFGS-B algorithm
        :param S_mat: list containing s_k vectors
        :return:
        :W matrix of the L-BFGS Hessian estimate
        :Y matrix containing the y_k vectors
        :S matrix containing the s_k vectors
        """
        Y = np.array([yk.getNdArray().ravel() for yk in Y_mat]).T
        S = np.array([sk.getNdArray().ravel() for sk in S_mat]).T
        W = np.concatenate((Y, theta*S), axis=1)
        return W, Y, S

    def max_step(self, bfgsb_mdl, bfgsb_dmodl, minBound, maxBound):
        """Method for determine maximum feasible step length"""
        n_mod = bfgsb_mdl.size
        x = bfgsb_mdl.getNdArray().ravel()
        l = minBound.getNdArray().ravel()
        u = maxBound.getNdArray().ravel()
        d = bfgsb_dmodl.getNdArray().ravel()
        max_alpha = 1e18
        for ii in range(n_mod):
            d1 = d[ii]
            if d1 < 0.0:
                d2 = l[ii] - x[ii]
                if  d2 >= 0.0:
                    max_alpha = 0.0
                elif d1 * max_alpha < d2:
                    max_alpha = d2 / (d1 + self.epsmch)
            elif d1 > 0.0:
                d2 = u[ii] - x[ii]
                if d2 <= 0.0:
                    max_alpha = zero
                elif d1 * max_alpha > d2:
                    max_alpha = d2 / (d1 + self.epsmch)
        return max_alpha

    def run(self, problem, verbose=False, restart=False):
        """
        Running LBFGSB solver
        :param problem: problem to be minimized
        :param verbose: verbosity flag [False]
        :param restart: restart previously crashed inversion [False]
        """
        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Getting model vector
        prblm_mdl = problem.get_model()

        # Obtaining bounds from problem
        minBound = problem.bounds.minBound
        maxBound = problem.bounds.maxBound
        if minBound is None:
            raise ValueError("Minimum bound vector must be provided when instantiating the problem object!")
        if maxBound is None:
            raise ValueError("Maximum bound vector must be provided when instantiating the problem object!")

        # Setting matrices (lists) of gradient differences and taken steps
        S_mat = []
        Y_mat = []
        # Numpy Arrays containing the matrices necessary for the L-BFGS-B internal optimization problems
        W = np.zeros((prblm_mdl.size,1))
        M = np.zeros((1, 1))

        if not restart:
            msg = 90 * "#" + "\n"
            if self.m_steps is not None:
                msg += "Limited-memory Broyden-Fletcher-Goldfarb-Shanno with Bounds (L-BFGS-B) algorithm log file\n"
                msg += "Maximum number of steps to be used for Hessian inverse estimation: %s \n" % self.m_steps
            else:
                msg += "Broyden-Fletcher-Goldfarb-Shanno with Bounds (BFGS-B) algorithm log file\n"
            # Printing restart folder
            msg += "Restart folder: %s\n" % self.restart.restart_folder
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace("log file", ""))
            if self.logger:
                self.logger.addToLog(msg)

            # Setting internal vectors (model, search direction, and previous gradient vectors)
            bfgsb_mdl = prblm_mdl.clone()
            bfgsb_dmodl = prblm_mdl.clone()
            bfgsb_dmodl.zero()
            bfgsb_grad0 = bfgsb_dmodl.clone()
            # Projecting initial guess within the bounded area
            if "bounds" in dir(problem):
                problem.bounds.apply(bfgsb_mdl)
            if prblm_mdl.isDifferent(bfgsb_mdl):
                # Model hit bounds
                msg = "\t!!!Initial guess outside of bounds. Projecting initial guess back into the feasible space.!!!"
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(bfgsb_mdl)
            # Other internal variables
            iiter = 0
            theta = 1.0
        else:
            # Retrieving parameters and vectors to restart the solver
            msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            k_steps = self.restart.retrieve_parameter("k_steps")
            theta = self.restart.retrieve_vector("theta")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            bfgsb_mdl = self.restart.retrieve_vector("bfgsb_mdl")
            bfgsb_dmodl = self.restart.retrieve_vector("bfgsb_dmodl")
            bfgsb_grad0 = self.restart.retrieve_vector("bfgsb_grad0")
            M = self.restart.retrieve_parameter("M")
            # Setting the model and residuals to avoid residual twice computation
            problem.set_model(bfgsb_mdl)
            prblm_mdl = problem.get_model()
            # Setting residual vector to avoid its unnecessary computation
            problem.set_residual(self.restart.retrieve_vector("prblm_res"))
            # Reading previously obtained S_mat and Y_mat
            k_vecs = min(k_steps,self.m_steps)
            if k_vecs > 0:
                for ii in range(k_vecs):
                    S_mat.append(self.restart.retrieve_vector("s_vec%s" % ii))
                    Y_mat.append(self.restart.retrieve_vector("y_vec%s" % ii))
                    W,_,_ = self.form_W(Y_mat, theta, S_mat)  # [Y theta*S]


        # Common variables unrelated to restart
        prev_mdl = prblm_mdl.clone().zero()
        bfgsb_mdl_cauchy = prblm_mdl.clone().zero() # Generalized Cauchy point vector
        # Precision of numbers from dot-product
        norm_mdl = prblm_mdl.norm()
        if isinstance(norm_mdl, np.float32):
            self.epsmch = np.finfo(np.float32).eps
        else:
            self.epsmch = np.finfo(np.float64).eps

        # Inversion loop
        while True:
            # Computing objective function
            obj0 = problem.get_obj(bfgsb_mdl)  # Compute objective function value
            prblm_res = problem.get_res(bfgsb_mdl)  # Compute residuals
            prblm_grad = problem.get_grad(bfgsb_mdl)  # Compute the gradient
            if iiter == 0:
                initial_obj_value = obj0  # For relative objective function value
                # Saving objective function value
                self.restart.save_parameter("obj_initial", initial_obj_value)
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj0,
                                       problem.get_rnorm(bfgsb_mdl),
                                       problem.get_gnorm(bfgsb_mdl),
                                       problem.get_fevals(),
                                       problem.get_gevals())
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0) or isnan(prblm_grad.norm()):
                    raise ValueError("Either gradient norm or objective function value NaN!")
            if prblm_grad.norm() == 0.:
                print("Gradient vanishes identically")
                break

            # Saving results
            self.save_results(iiter, problem, force_save=False)
            # Saving current inverted model
            prev_mdl.copy(prblm_mdl)
            # grad0 = grad
            bfgsb_grad0.copy(prblm_grad)

            # Compute the new search direction by finding Cauchy point and solving subspace minimization problem
            # [xc, c] = get_cauchy_point(x, g, l, u, theta, W, M);
            c = self.get_cauchy_point(bfgsb_mdl_cauchy, bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound, W, M, theta)
            self.stepper.alpha = 1.0
            if iiter == 0:
                bfgsb_dmodl.copy(bfgsb_mdl_cauchy)
                bfgsb_dmodl.scaleAdd(bfgsb_mdl, 1.0, -1.0)
                self.stepper.alpha = min(1.0, 1.0 / (bfgsb_dmodl.norm() + self.epsmch))
            else:
                free_var = self.subspace_min(bfgsb_dmodl, bfgsb_mdl, prblm_grad, minBound, maxBound, bfgsb_mdl_cauchy, c, W, M, theta)
                if not free_var:
                    bfgsb_dmodl.copy(bfgsb_mdl_cauchy)
                    bfgsb_dmodl.scaleAdd(bfgsb_mdl,1.0,-1.0)
                    self.stepper.alpha = min(1.0, 1.0 / (bfgsb_dmodl.norm() + self.epsmch))
            self.stepper.alpha_max = self.max_step(bfgsb_mdl, bfgsb_dmodl, minBound, maxBound)
            if self.stepper.alpha > self.stepper.alpha_max:
                self.stepper.alpha = self.stepper.alpha_max
            dphi = prblm_grad.dot(bfgsb_dmodl)
            if dphi > 0.0:
                msg = "Current search direction not a descent one (dphi/dalpha|dmod=%.5e)" % dphi
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break
            else:
                alpha, success = self.stepper.run(problem, bfgsb_mdl, bfgsb_dmodl, self.logger)
            if not success:
                msg = "Stepper couldn't find a proper step size, will terminate solver"
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break

            obj1 = problem.get_obj(bfgsb_mdl)  # Compute objective function value
            # Redundant tests on verifying convergence
            if obj0 <= obj1:
                msg = "Objective function at new point greater or equal than previous one: obj_fun_old=%s obj_fun_new=%s\n" \
                      "Stopping inversion!" % (obj0, obj1)
                if self.logger:
                    self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break

            # Compute new gradient
            prblm_grad = problem.get_grad(bfgsb_mdl)

            ###########################################
            # Update the LBFGS-B matrices
            # Gradient difference
            y_tmp = prblm_grad.clone()
            y_tmp.scaleAdd(bfgsb_grad0,1.0,-1.0)
            # Taken step
            s_tmp = bfgsb_mdl.clone()
            s_tmp.scaleAdd(prev_mdl,1.0,-1.0)

            # Checking curvature condition of equation 3.9 in "A LIMITED MEMORY ALGORITHM FOR BOUND
            # CONSTRAINED OPTIMIZATION" by Byrd et al. (1995)
            if y_tmp.dot(s_tmp) > self.epsmch*y_tmp.dot(y_tmp) and y_tmp.dot(s_tmp) > 0.0:
                msg = "Updating current Hessian estimate"
                if self.logger:
                    self.logger.addToLog(msg)
                if len(Y_mat) < self.m_steps:
                    # Appending new y and s vectors
                    Y_mat.append(y_tmp)
                    S_mat.append(s_tmp)
                else:
                    # Removing first column and updating last one
                    Y_mat[:-1] = Y_mat[1:]
                    Y_mat[-1] = y_tmp
                    S_mat[:-1] = S_mat[1:]
                    S_mat[-1] = s_tmp
                if iiter == 1 and len(Y_mat) == 2:
                    # Removing update from iteration 0
                    Y_mat.pop(0)
                    S_mat.pop(0)
                # Number of steps currently within L-BFGS matrix
                k_steps = len(Y_mat)
                if k_steps > 0:
                    theta = y_tmp.dot(y_tmp) / y_tmp.dot(s_tmp)
                    W, Y, S = self.form_W(Y_mat, theta, S_mat)  # [Y theta*S]
                    A = np.matmul(Y.T, S)
                    StS = np.matmul(S.T, S)
                    L = np.tril(A, -1)
                    D = -np.diag(np.diag(A))
                    M1 = np.concatenate((D, L.T), axis=1)
                    M2 = np.concatenate((L, theta * StS), axis=1)
                    MM = np.concatenate((M1, M2), axis=0)
                    M = np.linalg.inv(MM)
                    del L, D, M1, M2, MM, A
                    # Saving Inverse Hessian estimate for restart
                    self.restart.save_vector("y_vec%s" % (k_steps - 1), y_tmp)
                    self.restart.save_vector("s_vec%s" % (k_steps - 1), s_tmp)
            else:
                msg = "Skipping L-BFGS update; negative curvature detected"
                if self.logger:
                    self.logger.addToLog(msg)
            ###########################################
            # Increasing iteration counter
            iiter = iiter + 1

            # Saving current model and previous search direction in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_parameter("theta", theta)
            self.restart.save_parameter("M", M)
            self.restart.save_parameter("k_steps", k_steps)
            self.restart.save_vector("bfgsb_mdl", bfgsb_mdl)
            self.restart.save_vector("bfgsb_dmodl", bfgsb_dmodl)
            self.restart.save_vector("bfgsb_grad0", bfgsb_grad0)
            # Saving data space vectors
            self.restart.save_vector("prblm_res", prblm_res)

            # iteration info
            msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                   obj1,
                                   problem.get_rnorm(bfgsb_mdl),
                                   problem.get_gnorm(bfgsb_mdl),
                                   problem.get_fevals(),
                                   problem.get_gevals())
            if verbose:
                print(msg)
            # Writing on log file
            if self.logger:
                self.logger.addToLog("\n" + msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1) or isnan(prblm_grad.norm()):
                raise ValueError("Either gradient norm or objective function value NaN!")
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break

        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        msg = 90 * "#" + "\n"
        if self.m_steps is not None:
            msg += "Limited-memory Broyden-Fletcher-Goldfarb-Shanno with Bounds (L-BFGS-B) algorithm log file end\n"
        else:
            msg += "Broyden-Fletcher-Goldfarb-Shanno with Bounds (BFGS-B) algorithm log file end\n"
        msg += 90 * "#" + "\n"
        if verbose:
            print(msg.replace("log file ", ""))
        if self.logger:
            self.logger.addToLog(msg)
        self.restart.clear_restart()


class MCMC(S.Solver):
    """Markov chain Monte Carlo sampling algorithm"""
    
    def __init__(self, **kwargs):
        """
        Constructor for MCMC Solver/Sampler:
        :param stopper: Stopper object to terminate sampling
        :param prop_distr: proposal distribution to be employed ["Uni","Gauss"]
        1) "Uni" = uniform distribution: provide max_step U~[-min_step,max_step]
        2) "Gauss" = Gaussian distribution: provide sigma N~[0,sigma]
        :param logger: Logger, object to save inversion information at runtime [None]
        """
        # Calling parent construction
        super(MCMC, self).__init__()
        # Defining stopper object
        self.stopper = kwargs.get("stopper")
        # Logger object to write on log file
        self.logger = kwargs.get("logger", None)
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # Proposal distribution parameters
        self.prop_dist = kwargs.get("prop_distr")
        if self.prop_dist == "Uni":
            self.max_step = kwargs.get("max_step")
            self.min_step = kwargs.get("min_step", -self.max_step)
        elif self.prop_dist == "Gauss":
            self.sigma = kwargs.get("sigma")
        else:
            raise ValueError("Not supported prop_distr")
        # print formatting
        self.iter_msg = "sample number = %s, log-obj = %.5e, resnorm = %.2e, feval = %d, acceptance rate = %2.5f%%, alpha = %1.5f"
        self.ndigits = self.stopper.zfill
        # Temperature Metropolis sampling algorithm (see, Monte Carlo sampling of
        # solutions to inverse problems by Mosegaard and Tarantola, 1995)
        # If not provided the likelihood is assumed to be passed to the run method
        self.T = kwargs.get("T", None)
        # Reject sample outside of bounds or project it onto them
        self.reject_bound = True
    
    def run(self, problem, verbose=False, restart=False):
        """Running MCMC solver/sampler"""
        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Checking if user is saving the sampled models
        if not self.save_model:
            msg = "WARNING! save_model=False! Running MCMC sampling method will not save accepted model samples!"
            print(msg)
            if self.logger:
                self.logger.addToLog(msg)
        
        if not restart:
            msg = 90 * "#" + "\n"
            msg += "Markov Chain Monte Carlo (MCMC) algorithm log file\n"
            msg += "Restart folder: %s\n" % self.restart.restart_folder
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace("log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
            prblm_mdl = problem.get_model()
            mcmc_mdl_cur = prblm_mdl.clone()
            
            # Other internal variables
            accepted = 1  # number of accepted samples
            iiter = 1  # number of tested point so far
        else:
            # Retrieving parameters and vectors to restart the solver
            msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
            if verbose:
                print(msg)
            if self.logger:
                self.logger.addToLog(msg)
            self.restart.read_restart()
            mcmc_mdl_cur = self.restart.retrieve_vector("mcmc_mdl_cur")
            accepted = self.restart.retrieve_parameter("accepted")
            iiter = self.restart.retrieve_parameter("iiter")
            # Setting the last accepted model
            problem.set_model(mcmc_mdl_cur)
            prblm_mdl = problem.get_model()
        
        # Common parameters
        mcmc_mdl_prop = prblm_mdl.clone()
        mcmc_dmodl = prblm_mdl.clone().zero()
        mcmc_mdl_check = prblm_mdl.clone()
        
        # Computing current objective function
        obj_current = problem.get_obj(mcmc_mdl_cur)  # Compute objective function value
        res_norm = problem.get_rnorm(mcmc_mdl_cur)
        # getting each objective function term if present
        obj_terms = problem.obj_terms if "obj_terms" in dir(problem) else None
        if not restart:
            # iteration info
            msg = self.iter_msg % (str(iiter).zfill(self.ndigits),
                                   np.log(obj_current),
                                   res_norm,
                                   problem.get_fevals(),
                                   100. * float(accepted) / iiter,
                                   1.0)
            if verbose:
                print(msg)
            # Writing on log file
            if self.logger:
                self.logger.addToLog("\n" + msg)
            # Check if objective function value is NaN
            if isnan(obj_current):
                raise ValueError("objective function value NaN!")
            self.save_results(iiter, problem, force_save=False)
        # Sampling loop
        while True:
            # Generate a candidate y from x according to the proposal distribution r(x_cur, x_prop)
            if self.prop_dist == "Uni":
                mcmc_dmodl.getNdArray()[:] = np.random.uniform(low=self.min_step, high=self.max_step,
                                                               size=mcmc_dmodl.shape)
            elif self.prop_dist == "Gauss":
                mcmc_dmodl.getNdArray()[:] = np.random.normal(scale=self.sigma, size=mcmc_dmodl.shape)
            # Compute a(x_cur, x_prop)
            mcmc_mdl_prop.copy(mcmc_mdl_cur)
            mcmc_mdl_prop.scaleAdd(mcmc_dmodl)
            # Checking if model parameters hit the bounds
            mcmc_mdl_check.copy(mcmc_mdl_prop)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(mcmc_mdl_check)
            if mcmc_mdl_prop.isDifferent(mcmc_mdl_check):
                msg = "\tModel hit provided bounds. Projecting onto them."
                if self.reject_bound:
                    msg = "\tModel hit provided bounds. Resampling proposed point."
                # Model hit bounds
                if self.logger:
                    self.logger.addToLog(msg)
                if self.reject_bound:
                    continue
                else:
                    mcmc_mdl_prop.copy(mcmc_mdl_check)
            
            obj_prop = problem.get_obj(mcmc_mdl_prop)
            # Check if objective function value is NaN
            if isnan(obj_prop):
                raise ValueError("objective function of proposed model is NaN!")
            if self.T:
                # Using Metropolis method assuming an objective function was passed
                alpha = 1.0
                if obj_prop > obj_current:
                    alpha = np.exp(-(obj_prop - obj_current) / self.T)
            else:
                # computing log acceptance ratio assuming likelihood function
                if obj_current > zero and obj_prop > zero:
                    alpha = min(1.0, obj_prop / obj_current)
                elif obj_current <= zero < obj_prop:
                    # condition to avoid zero/zero
                    alpha = 0.
                else:
                    # condition to avoid division by zero
                    alpha = 1.
            
            # Increase counter of tested samples
            iiter += 1
            
            # Accept the x_prop with probability a
            if np.random.uniform() <= alpha:
                # accepted proposed model
                mcmc_mdl_cur.copy(mcmc_mdl_prop)
                obj_current = deepcopy(obj_prop)
                obj_terms = deepcopy(problem.obj_terms) if "obj_terms" in dir(problem) else None
                res_norm = problem.get_rnorm(mcmc_mdl_cur)
                accepted += 1  # Increase counter of accepted samples
            
            # iteration info
            msg = self.iter_msg % (str(iiter).zfill(self.ndigits),
                                   np.log(obj_current),
                                   res_norm,
                                   problem.get_fevals(),
                                   100. * float(accepted) / iiter,
                                   alpha)
            if verbose:
                print(msg)
            # Writing on log file
            if self.logger:
                self.logger.addToLog("\n" + msg)
            
            # Saving sampled point
            self.save_results(iiter, problem, model=mcmc_mdl_cur, obj=obj_current, obj_terms=obj_terms, force_save=True,
                              force_write=True)
            
            # saving inversion vectors and parameters for restart
            self.restart.save_vector("mcmc_mdl_cur", mcmc_mdl_cur)
            self.restart.save_parameter("accepted", accepted)
            self.restart.save_parameter("iiter", iiter)
            
            # Checking stopping criteria
            if self.stopper.run(problem, iiter, verbose=verbose):
                break
        
        msg = 90 * "#" + "\n"
        msg += "Markov Chain Monte Carlo (MCMC) algorithm algorithm log file end\n"
        msg += 90 * "#" + "\n"
        if verbose:
            print(msg.replace("log file ", ""))
        if self.logger:
            self.logger.addToLog(msg)
        self.restart.clear_restart()
        return
